import random
from typing import List, Literal, Optional, Union, cast

from pylabrobot.machines import Machine
from pylabrobot.resources import (
  Coordinate,
  Plate,
  PlateCarrier,
  PlateHolder,
  Resource,
  ResourceNotFoundError,
  Rotation,
)
from pylabrobot.serializer import deserialize, serialize

from .backend import IncubatorBackend


class NoFreeSiteError(Exception):
  pass


class Incubator(Machine, Resource):
  def __init__(
    self,
    backend: IncubatorBackend,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    racks: List[PlateCarrier],
    loading_tray_location: Coordinate,
    rotation: Optional[Rotation] = None,
    category: Optional[str] = None,
    model: Optional[str] = None,
  ):
    Machine.__init__(self, backend=backend)
    self.backend: IncubatorBackend = backend  # fix type
    Resource.__init__(
      self,
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      rotation=rotation,
      category=category,
      model=model,
    )
    self.loading_tray = PlateHolder(
      name=self.name + "_tray", size_x=127.76, size_y=85.48, size_z=0, pedestal_size_z=0
    )
    self.assign_child_resource(self.loading_tray, location=loading_tray_location)

    self._racks = racks
    for rack in self._racks:
      self.assign_child_resource(rack, location=None)

  @property
  def racks(self) -> List[PlateCarrier]:
    return self._racks

  async def setup(self, **backend_kwargs):
    await super().setup()
    await self.backend.set_racks(self._racks)

  def get_num_free_sites(self) -> int:
    return sum(len(rack.get_free_sites()) for rack in self._racks)

  def get_site_by_plate_name(self, plate_name: str) -> PlateHolder:
    for rack in self._racks:
      for site in rack.sites.values():
        if site.resource is not None and site.resource.name == plate_name:
          return site
    raise ResourceNotFoundError(f"Plate {plate_name} not found in incubator '{self.name}'")

  async def fetch_plate_to_loading_tray(self, plate_name: str, **backend_kwargs) -> Plate:
    """Fetch a plate from the incubator and put it on the loading tray."""

    site = self.get_site_by_plate_name(plate_name)
    plate = site.resource
    assert plate is not None
    await self.backend.fetch_plate_to_loading_tray(plate, **backend_kwargs)
    plate.unassign()
    self.loading_tray.assign_child_resource(plate)
    return plate

  def _find_available_sites_sorted(self, plate: Plate) -> List[PlateHolder]:
    """Find all sites that are free and fit the plate, sorted by size."""

    def _plate_height(p: Plate):
      if p.has_lid():
        # TODO: we can use plr nesting height
        # lid.location.z + lid.get_anchor(z="t").z
        return p.get_size_z() + 3
      return p.get_size_z()

    available = [
      site
      for rack in self._racks
      for site in rack.get_free_sites()
      if site.get_size_z() >= _plate_height(plate)
    ]
    if len(available) == 0:
      raise NoFreeSiteError(
        f"No free site found in incubator '{self.name}' for plate '{plate.name}'"
      )
    return sorted(available, key=lambda site: site.get_size_z())

  def find_smallest_site_for_plate(self, plate: Plate) -> PlateHolder:
    return self._find_available_sites_sorted(plate)[0]

  def find_random_site(self, plate: Plate) -> PlateHolder:
    return random.choice(self._find_available_sites_sorted(plate))

  async def take_in_plate(
    self, site: Union[PlateHolder, Literal["random", "smallest"]], **backend_kwargs
  ):
    """Take a plate from the loading tray and put it in the incubator."""

    plate = self.loading_tray.resource
    if not isinstance(plate, Plate):
      raise ResourceNotFoundError(f"No plate on the loading tray of incubator '{self.name}'")

    if site == "random":
      site = self.find_random_site(plate)
    elif site == "smallest":
      site = self.find_smallest_site_for_plate(plate)
    elif isinstance(site, PlateHolder):
      if site not in self._find_available_sites_sorted(plate):
        raise ValueError(f"Site {site.name} is not available for plate {plate.name}")
    else:
      raise ValueError(f"Invalid site: {site}")
    await self.backend.take_in_plate(plate, site, **backend_kwargs)
    plate.unassign()
    site.assign_child_resource(plate)

  async def set_temperature(self, temperature: float):
    """Set the temperature of the incubator in degrees Celsius."""
    return await self.backend.set_temperature(temperature)

  async def get_temperature(self) -> float:
    return await self.backend.get_temperature()

  async def open_door(self):
    return await self.backend.open_door()

  async def close_door(self):
    return await self.backend.close_door()

  async def start_shaking(self, frequency: float = 1.0):
    await self.backend.start_shaking(frequency=frequency)

  async def stop_shaking(self):
    await self.backend.stop_shaking()

  def summary(self) -> str:
    def create_pretty_table(header, *columns) -> str:
      col_widths = [
        max(len(str(item)) for item in [header[i]] + list(columns[i])) for i in range(len(header))
      ]

      def format_row(row, border="|") -> str:
        return (
          f"{border} "
          + " | ".join(f"{str(row[i]).ljust(col_widths[i])}" for i in range(len(row)))
          + f" {border}"
        )

      def separator_line(cross: str = "+", line: str = "-") -> str:
        return cross + cross.join(line * (width + 2) for width in col_widths) + cross

      table = []
      table.append(separator_line())  # Top border
      table.append(format_row(header))
      table.append(separator_line())  # Header separator
      for row in zip(*columns):
        table.append(format_row(row))
      table.append(separator_line())  # Bottom border
      return "\n".join(table)

    header = [f"Rack {i}" for i in range(len(self._racks))]
    sites = [
      [site.resource.name if site.resource else "<empty>" for site in reversed(rack.sites.values())]
      for rack in self._racks
    ]
    return create_pretty_table(header, *sites)

  def serialize(self):
    return {
      **Machine.serialize(self),
      **Resource.serialize(self),
      "backend": self.backend.serialize(),
      "racks": [rack.serialize() for rack in self._racks],
      "loading_tray_location": serialize(self.loading_tray.location),
    }

  @classmethod
  def deserialize(cls, data: dict, allow_marshal: bool = False):
    backend = IncubatorBackend.deserialize(data.pop("backend"))
    return cls(
      backend=backend,
      name=data["name"],
      size_x=data["size_x"],
      size_y=data["size_y"],
      size_z=data["size_z"],
      racks=[PlateCarrier.deserialize(rack) for rack in data["racks"]],
      loading_tray_location=cast(Coordinate, deserialize(data["loading_tray_location"])),
      rotation=Rotation.deserialize(data["rotation"]),
      category=data["category"],
      model=data["model"],
    )



import time
import typing

from pylabrobot.io.hid import HID


class InhecoTECControlBox:
  def __init__(
    self,
    vid=0x03EB,
    pid=0x2023,
    serial_number=None,
  ):
    self.io = HID(
      human_readable_device_name="Inheco Control Box", vid=vid, pid=pid, serial_number=serial_number
    )

  async def setup(self):
    """
    If io.setup() fails, ensure that libusb drivers were installed as per docs.
    """
    await self.io.setup()

  async def stop(self):
    await self.io.stop()

  @typing.no_type_check
  def _generate_packets(self, msg):
    """Generate packets for the given message.

    Splits the message into packets of 7 bytes each. The last packet contains the checksum.
    """

    ch_array1 = [None] * 128
    report_buffer = [0] * 9
    crc = 161
    for i, char in enumerate(msg):
      ch_array1[i] = char if ord(char) <= ord("`") or char == "#" else chr(ord(char) - 32)
      if ch_array1[i] != "#":
        crc = self._crc8(ord(ch_array1[i]), crc=crc)
    ch_array1[len(msg)] = "w" if crc in (35, 0) else chr(crc)
    num1 = len(msg) + 1
    if num1 < 2:
      report_buffer[0] = 0
    if num1 <= 8:
      for i in range(num1):
        report_buffer[i + 1] = ord(ch_array1[i])
      return [report_buffer]
    else:
      num2 = (num1 - 1) // 7  # number of packets
      command = []
      for index1 in range(num2 + 1):
        for index2 in range(7):
          if index2 + index1 * 7 >= num1:
            report_buffer[index2 + 1] = 0
          else:
            report_buffer[index2 + 1] = ord(ch_array1[index2 + index1 * 7])
        if index1 != num2:
          report_buffer[8] = ord("#")
        else:
          if (num2 + 1) * 7 + 1 == num1:
            report_buffer[8] = ord(ch_array1[num1 - 1])
            if report_buffer[8] > 96:
              report_buffer[8] -= 32
          else:
            report_buffer[8] = 0
        command.append(report_buffer)
      return command

  def _crc8(self, data, crc: int) -> int:
    """Meme crc8 implementation"""
    num = 8
    while num > 0:
      if ((data ^ crc) & 1) == 1:
        crc ^= 24
        crc >>= 1
        crc |= 128
      else:
        crc >>= 1
      data >>= 1
      num -= 1
    return crc

  async def _read_until_end(self, timeout: int) -> str:
    """Read until a packet ends with a \\x00 byte. May read multiple packets."""
    start = time.time()
    response = b""
    while time.time() - start < timeout:
      packet = await self.io.read(64, timeout=timeout)
      if packet is not None and packet != b"":
        if packet.endswith(b"\x00"):
          response += packet.rstrip(b"\x00")  # strip trailing \x00's
          break
        elif packet.endswith(b"#"):
          response += packet[:-1]
          continue
        else:
          # I have never seen this happen, commands always end with \x00 or '#'
          print("weird packet, please report", packet)
          response += packet

    return response.decode("unicode_escape")

  async def _read_response(self, command: str, timeout: int = 60) -> str:
    """Read the response for a given command.

    "The MTC/STC replies to the first four characters of every command with a modified echo. The
    modification changes the capitals of the commands to small letters. i.e. the reply to 5ASE1
    is 5ase0. Therefore it is easy to identify correct answers to the commands. This feature may
    increase integrity of the communication."
    """

    start = time.time()
    while time.time() - start < timeout:
      response = await self._read_until_end(timeout=int(timeout - (time.time() - start)))

      if response[:4] == command[:4].lower():
        return response

    raise TimeoutError("Timeout while waiting for response from device.")

  async def send_command(self, command: str, timeout: int = 3):
    """Send a command to the device and return the response"""
    packets = self._generate_packets(command)
    for packet in packets:
      await self.io.write(bytes(packet[1:]), report_id=bytes(packet[0]))

    response = await self._read_response(command, timeout=timeout)

    if response[4] != "0":
      raise RuntimeError(f"Error response from device: {response}")

    return response[5:-1]  # cut off command, error status, and final checksum byte

  async def set_touchscreen(self, active: bool):
    await self.send_command(f"0ADD{1 if active else 0}")
    await self.reset_action_display()

  async def reset_action_display(self):
    return await self.send_command("0ASD")



"""
Backend for Inheco Incubator Shaker Stack machine.

This module implements a fully asynchronous serial communication backend for
Inheco Incubator/Shaker instruments (e.g., Inheco MP/DWP with or without shaker).

Features:
- Auto-discovery of Inheco devices by VID:PID (error if more than one is found).
- Validation of DIP switch ID.
- Automatic identification of units in stack.
- Complete command/response layer with legacy CRC-8 and async-safe I/O.
- Structured firmware error reporting via InhecoError and contextual snapshotting.
- High-level API for temperature control, drawer handling, and shaking functions.
- Protocol-conformant parsing for EEPROM, sensor, and status commands.
"""

import asyncio
import logging
import sys
from functools import wraps
from typing import Awaitable, Callable, Dict, List, Literal, Optional, TypeVar, cast

from pylabrobot.io.serial import Serial
from pylabrobot.machines.machine import MachineBackend

if sys.version_info < (3, 10):
  from typing_extensions import Concatenate, ParamSpec
else:
  from typing import Concatenate, ParamSpec
try:
  import serial

  HAS_SERIAL = True
except ImportError as e:
  HAS_SERIAL = False
  _SERIAL_IMPORT_ERROR = e


P = ParamSpec("P")
R = TypeVar("R")


class InhecoError(RuntimeError):
  """Represents an Inheco firmware-reported error."""

  def __init__(self, command: str, code: str, message: str):
    super().__init__(f"{command} failed with error {code}: {message}")
    self.command: str = command
    self.code: str = code
    self.message: str = message


_REF_FLAG_NAMES: Dict[int, str] = {
  # Heater (0-15) — names per manual's heater flags table (subset shown here)
  0: "H_WARN_WarmUp_TIME",
  1: "H_WARN_BoostCoolDown_TIME",
  2: "H_WARN_StartState_LIMIT_Up_TEMP_S2",
  3: "H_WARN_StartState_LIMIT_Up_TEMP_S3",
  4: "H_WARN_StartStateBoost_LIMIT_UpDown_TEMP_S3",
  5: "H_WARN_StableState_LIMIT_UpDown_TEMP_S2",
  6: "H_WARN_StableState_LIMIT_UpDown_TEMP_S3",
  7: "H_WARN_DELTA_TEMP_S1_S2",
  8: "H_ERR_DELTA_TEMP_S1_S2",
  9: "H_WARN_StartStateBoost_LIMIT_UpDown_TEMP_S2",
  10: "H_WARN_WaitStable_LIMIT_TEMP_S1",
  11: "H_WARN_WaitStable_LIMIT_TEMP_S2",
  12: "H_WARN_WaitStable_LIMIT_TEMP_S3",
  13: "H_ERR_S2_NTC_NotConnected",
  14: "H_ERR_S3_NTC_NotConnected",
  15: "H_WARN_DELTA_TEMP_S1_S3",
  # Shaker (16-26) — names per manual's shaker flag set (page 39)
  16: "S_WARN_MotorCurrentLimit",
  17: "S_WARN_TargetSpeedTimeout",
  18: "S_WARN_PositionTimeout",
  19: "S_ERR_MotorTemperatureLimit",
  20: "S_ERR_TargetSpeedDeviation",
  21: "S_ERR_HomeSensorTimeout",
  22: "S_ERR_MotorDriverFault",
  23: "S_ERR_EncoderSignalLost",
  24: "S_ERR_AmplitudeOutOfRange",
  25: "S_ERR_VibrationExcessive",
  26: "S_ERR_InternalTimeout",
  # 27-31 reserved
}

FIRMWARE_ERROR_MAP: Dict[int, str] = {
  0: "Msg Ok",
  1: "Reset detected",
  2: "Invalid command",
  3: "Invalid operand",
  4: "Protocol error",
  5: "Reserved",
  6: "Timeout from Device",
  7: "Device not initialized",
  8: "Command not executable",
  9: "Drawer not in end position",
  10: "Unexpected Labware Status",
  13: "Drawer DWP not perfectly closed (NTC not connected)",
  14: "Floor ID error",
  15: "Timeout sub device",
}

InhecoIncubatorUnitType = Literal[
  "incubator_mp", "incubator_shaker_mp", "incubator_dwp", "incubator_shaker_dwp", "unknown"
]


def requires_incubator_shaker(
  func: Callable[Concatenate["InhecoIncubatorShakerStackBackend", P], Awaitable[R]],
) -> Callable[Concatenate["InhecoIncubatorShakerStackBackend", P], Awaitable[R]]:
  @wraps(func)
  async def wrapper(
    self: "InhecoIncubatorShakerStackBackend", *args: P.args, **kwargs: P.kwargs
  ) -> R:
    name = getattr(func, "__name__", func.__class__.__name__)
    stack_index = cast(int, kwargs.get("stack_index", 0))
    incubator_type = self.unit_composition[stack_index]

    if "shaker" not in incubator_type:
      raise RuntimeError(f"{name}() requires a shaker-capable model (got {incubator_type!r}).")

    return await func(self, *args, **kwargs)

  return wrapper


class InhecoIncubatorShakerStackBackend(MachineBackend):
  """Interface for Inheco Incubator Shaker stack machines.

  Handles:
    - USB/serial connection setup via VID/PID
    - DIP switch ID verification
    - Message framing, CRC generation
    - Complete async read/write of firmware responses
    - Binary-safe parsing and error mapping
  """

  # === Logging utility ===

  # === Constructor ===

  def __init__(
    self,
    dip_switch_id: int = 2,
    port: Optional[str] = None,
    write_timeout: float = 5.0,
    read_timeout: float = 10.0,
    vid: int = 0x0403,
    pid: int = 0x6001,
  ):
    if not HAS_SERIAL:
      raise RuntimeError(
        "pyserial is not installed. Install with: pip install pylabrobot[serial]. "
        f"Import error: {_SERIAL_IMPORT_ERROR}"
      )
    super().__init__()

    self.logger = logging.LoggerAdapter(
      logging.getLogger(__name__),
      {"dip_switch_id": dip_switch_id},
    )

    # Core state
    self.dip_switch_id = dip_switch_id

    self.io = Serial(
      human_readable_device_name="Inheco Incubator Shaker",
      port=port,
      vid=vid,
      pid=pid,
      baudrate=19_200,
      bytesize=serial.EIGHTBITS,
      parity=serial.PARITY_NONE,
      stopbits=serial.STOPBITS_ONE,
      timeout=0,
      write_timeout=1,
    )

    # Communication timeouts (defaults for all units in stack)
    self.write_timeout = write_timeout
    self.read_timeout = read_timeout

    # Cached state (stack level)
    self.setup_finished = False
    self.max_temperature = 85.0  # safe default
    self.unit_composition: List[
      InhecoIncubatorUnitType
    ] = []  # e.g. ["incubator_mp", "incubator_shaker_dwp", ...]

    self._send_command_lock = asyncio.Lock()

  @property
  def number_of_connected_units(self) -> int:
    """Return the number of connected units in the stack."""
    return len(self.unit_composition)

  def __repr__(self):
    return (
      f"<InhecoIncubatorShakerBackend (VID:PID={self.io._vid}:{self.io._pid}, "
      + f"DIP={self.dip_switch_id}) at {self.io.port}>"
    )

  async def setup(self, port: Optional[str] = None):
    """
    Detect and connect to the Inheco machine stack.
    Discover Inheco device via VID:PID (0403:6001) and verify DIP switch ID.
    """

    # --- Establish serial connection ---
    await self.io.setup()
    self.io.dtr = False
    self.io.rts = False

    try:  # --- Verify DIP switch ID via RTS ---
      probe = self._build_message("RTS", stack_index=0)
      await self.io.write(probe)
      resp = await self._read_full_response(timeout=5.0)

      expected_hdr = (0xB0 + self.dip_switch_id) & 0xFF
      if resp[0] != expected_hdr:
        raise ValueError("Unexpected header")

    except Exception as e:
      # Capture current IO reference before marking disconnected
      msg = (
        f"Device on {self.io._port} failed DIP switch verification (expected ID="
        f"{self.dip_switch_id}). Please verify the DIP switch setting or wiring."
      )
      self.logger.error(msg, exc_info=e)

      # --- Fail-safe teardown ---
      try:
        await self.io.stop()
        self.logger.debug("Closed serial connection on %s", self.io.port)
      except Exception as close_err:
        self.logger.warning(
          "Failed to close serial port cleanly on %s: %s",
          self.io._port,
          close_err,
        )
      raise RuntimeError(msg) from e

    else:
      # Connection verified and active
      self.logger.log(
        logging.INFO,
        f"Connected to Inheco machine at {self.io.port} (DIP={self.dip_switch_id})",
      )

    # --- Cache stack-level state ---
    number_of_connected_units = await self.request_number_of_connected_machines(stack_index=0)

    self.unit_composition = []

    for unit_index in range(number_of_connected_units):
      inc_type = await self.request_incubator_type(stack_index=unit_index)
      self.unit_composition.append(inc_type)

      await self.initialize(stack_index=unit_index)

    self.setup_finished = True

    self.logger.info(
      "Connected to Inheco Incubator Shaker Stack on %s\n"
      "DIP switch ID of bottom unit: %s\n"
      "Number of connected units: %s\n"
      "Unit composition: %s",
      self.io.port,
      self.dip_switch_id,
      self.number_of_connected_units,
      self.unit_composition,
    )

  async def stop(self):
    """Close serial connection & stop all active units in the stack."""

    for unit_index in range(self.number_of_connected_units):
      temp_status = await self.is_temperature_control_enabled(stack_index=unit_index)

      if temp_status:
        print(f"Stopping temperature control on unit {unit_index}...")
        await self.stop_temperature_control(stack_index=unit_index)

      shake_status = await self.is_shaking_enabled(stack_index=unit_index)

      if shake_status:
        print(f"Stopping shaking on unit {unit_index}...")
        await self.stop_shaking(stack_index=unit_index)

      await self.close(stack_index=unit_index)

    await self.io.stop()

  # === Low-level I/O ===

  async def _read_full_response(self, timeout: float) -> bytes:
    """Read a complete Inheco response frame asynchronously."""
    loop = asyncio.get_event_loop()
    start = loop.time()
    buf = bytearray()
    expected_hdr = (0xB0 + self.dip_switch_id) & 0xFF

    def has_complete_tail(b: bytearray) -> bool:
      # Valid frame ends with: [hdr][0x20-0x2F][0x60]
      return len(b) >= 3 and b[-1] == 0x60 and b[-3] == expected_hdr and 0x20 <= b[-2] <= 0x2F

    while True:
      chunk = await self.io.read(16)
      if len(chunk) > 0:
        buf.extend(chunk)
        if has_complete_tail(buf):
          self.logger.debug("RECV response: %s", buf.hex(" "))
          return bytes(buf)

      if loop.time() - start > timeout:
        raise TimeoutError(f"Timed out waiting for complete response (so far: {buf.hex(' ')})")

      await asyncio.sleep(0.005)

  # === Encoding / Decoding ===

  def _crc8_legacy(self, data: bytearray) -> int:
    """Compute legacy CRC-8 used by Inheco devices."""
    crc = 0xA1
    for byte in data:
      d = byte
      for _ in range(8):
        if (d ^ crc) & 1:
          crc ^= 0x18
          crc >>= 1
          crc |= 0x80
        else:
          crc >>= 1
        d >>= 1
    return crc & 0xFF

  def _build_message(self, command: str, stack_index: int = 0) -> bytes:
    """Construct a full binary message with header and CRC."""
    if not (0 <= stack_index <= 5):
      raise ValueError("stack_index must be between 0 and 5")
    cmd = f"T0{stack_index}{command}".encode("ascii")
    length = len(cmd) + 3
    address = (0x30 + self.dip_switch_id) & 0xFF
    proto = (0xC0 + len(cmd)) & 0xFF
    message = bytearray([length, address, proto]) + cmd
    crc = self._crc8_legacy(message)
    return bytes(message + bytearray([crc]))

  def _is_report_command(self, command: str) -> bool:
    """Return True if command is a 'Report' type (starts with 'R')."""
    return len(command) > 0 and command[0].upper() == "R"

  # === Response parsing ===

  def _parse_response_binary_safe(self, resp: bytes) -> dict:
    """Parse Inheco response frames safely (binary & multi-segment)."""
    if len(resp) < 3:
      raise ValueError("Incomplete response")

    expected_hdr = (0xB0 + self.dip_switch_id) & 0xFF

    # Trim any leading junk before first valid header
    try:
      start_idx = resp.index(bytes([expected_hdr]))
      frame = resp[start_idx:]
    except ValueError:
      return {
        "device": None,
        "error_code": None,
        "ok": False,
        "data": "",
        "raw_data": resp,
      }

    # Validate tail
    if len(frame) < 3 or frame[-1] != 0x60:
      return {
        "device": expected_hdr - 0xB0,
        "error_code": None,
        "ok": False,
        "data": "",
        "raw_data": frame,
      }

    err_byte = frame[-2]
    err_code = err_byte - 0x20 if 0x20 <= err_byte <= 0x2F else None

    # Extract data between headers
    data_blocks = []
    i = 1  # start after first header
    while i < len(frame) - 3:
      try:
        next_hdr = frame.index(bytes([expected_hdr]), i)
      except ValueError:
        next_hdr = len(frame) - 3
      if next_hdr > i:
        data_blocks.append(frame[i:next_hdr])
      i = next_hdr + 1
      if next_hdr >= len(frame) - 3:
        break

    raw_data = b"".join(data_blocks)
    try:
      ascii_data = raw_data.decode("ascii").strip("\x00")
    except UnicodeDecodeError:
      ascii_data = raw_data.hex()

    return {
      "device": expected_hdr - 0xB0,
      "error_code": err_code,
      "ok": (err_code == 0),
      "data": ascii_data,
      "raw_data": raw_data,
    }

  def _is_error_tail(self, resp: bytes) -> bool:
    """Return True if the response ends in an explicit firmware error tail."""
    expected_hdr = (0xB0 + self.dip_switch_id) & 0xFF
    return len(resp) >= 3 and resp.endswith(bytes([expected_hdr, 0x28, 0x60]))

  # === Command Layer ===

  async def send_command(
    self,
    command: str,
    *,
    delay: float = 0.2,
    write_timeout: Optional[float] = None,
    read_timeout: Optional[float] = None,
    stack_index: int = 0,
  ) -> str:
    """Send a framed command and return parsed response or raise InhecoError."""

    async with self._send_command_lock:
      # Use global default if not overridden
      w_timeout = write_timeout or self.write_timeout
      msg = self._build_message(command, stack_index=stack_index)
      self.logger.debug("SEND command: %s (write_timeout=%s)", msg.hex(" "), w_timeout)

      await asyncio.wait_for(self.io.write(msg), timeout=w_timeout)
      await asyncio.sleep(delay)

      response = await self._read_full_response(timeout=read_timeout or self.read_timeout)
      if not response:
        raise TimeoutError(f"No response from device for command: {command}")

      if self._is_error_tail(response):
        tail_err = response[-2] - 0x20
        code = f"E{tail_err:02d}"
        message = FIRMWARE_ERROR_MAP.get(tail_err, "Unknown firmware error")
        raise InhecoError(command, code, message)

      parsed = self._parse_response_binary_safe(response)
      if not parsed["ok"]:
        code = f"E{parsed.get('error_code', 0):02d}"
        message = FIRMWARE_ERROR_MAP.get(parsed.get("error_code", 0), "Unknown firmware error")
        raise InhecoError(command, code, message)

      return str(parsed["data"])

  # === Public high-level API ===

  # Querying Machine State #
  async def request_firmware_version(self, stack_index: int) -> str:
    """EEPROM request: Return the firmware version string."""
    resp = await self.send_command("RFV0", stack_index=stack_index)
    return resp

  async def request_serial_number(self, stack_index: int) -> str:
    """EEPROM request: Return the device serial number."""
    resp = await self.send_command("RFV2", stack_index=stack_index)
    return resp

  async def request_last_calibration_date(self, stack_index: int) -> str:
    """EEPROM request"""
    resp = await self.send_command("RCM", stack_index=stack_index)
    return resp[:10]

  async def request_machine_allocation(self, layer: int = 0, stack_index: int = 0) -> dict:
    """
    Report which device slots are occupied on a given layer (firmware 'RDAx,0').

    Args:
      layer: Layer index (0-7). Default = 0.

    Returns:
      {
        "layer": int,
        "slot_mask": int,       # e.g. 7
        "slot_mask_bin": str,   # e.g. "0b0000000000000111"
        "slots_connected": list[int]  # e.g. [0, 1, 2]
      }

    Notes:
      Each bit in `slot_mask` represents one of 16 possible device slots:
      bit=1 means a device is connected; bit=0 means empty.
    """
    if not (0 <= layer <= 7):
      raise ValueError(f"Layer must be between 0 and 7, got {layer}")

    resp = await self.send_command(f"RDA{layer},0", stack_index=stack_index)
    resp_str = str(resp).strip()
    slot_mask = int(resp_str)
    slot_mask_bin = f"0b{slot_mask:016b}"

    slots_connected = [i for i in range(16) if (slot_mask >> i) & 1]

    return {
      "layer": layer,
      "slot_mask": slot_mask,
      "slot_mask_bin": slot_mask_bin,
      "slots_connected": slots_connected,
    }

  async def request_number_of_connected_machines(self, layer: int = 0, stack_index: int = 0) -> int:
    """
    Report the number of connected Inheco devices on a layer (RDAx,1).

    Args:
      layer: Layer index (0-7). Default = 0.

    Returns:
      Number of connected devices (0-16).

    Example:
      Response "3" → 3 connected devices on that layer.
    """
    if not (0 <= layer <= 7):
      raise ValueError(f"Layer must be 0-7, got {layer}")

    resp = await self.send_command(f"RDA{layer},1", stack_index=stack_index)
    return int(resp.strip())

  async def request_labware_detection_threshold(self, stack_index: int) -> int:
    """EEPROM request"""
    resp = await self.send_command("RDM", stack_index=stack_index)
    return int(resp)

  async def request_incubator_type(self, stack_index: int) -> InhecoIncubatorUnitType:
    """Return a descriptive string of the incubator/shaker configuration."""

    incubator_type_dict = {
      "0": "incubator_mp",  # no shaker
      "1": "incubator_shaker_mp",
      "2": "incubator_dwp",  # no shaker
      "3": "incubator_shaker_dwp",
    }
    resp = await self.send_command("RTS", stack_index=stack_index)
    return incubator_type_dict.get(resp, "unknown")  # type: ignore[return-value]

  async def request_plate_in_incubator(self, stack_index: int) -> bool:
    """Sensor command:"""
    resp = await self.send_command("RLW", stack_index=stack_index)
    return resp == "1"

  async def request_operation_time_in_hours(self, stack_index: int) -> int:
    """EEPROM request"""
    resp = await self.send_command("RDC1", stack_index=stack_index)
    return int(resp)

  async def request_drawer_cycles_performed(self, stack_index: int) -> int:
    """EEPROM request"""
    resp = await self.send_command("RDC2", stack_index=stack_index)
    return int(resp)

  async def request_is_initialized(self, stack_index: int) -> bool:
    """EEPROM request"""
    resp = await self.send_command("REE", stack_index=stack_index)
    return resp in {"0", "2"}

  async def request_plate_status_known(self, stack_index: int) -> bool:
    """EEPROM request"""
    resp = await self.send_command("REE", stack_index=stack_index)
    return resp in {"0", "1"}

  async def request_thermal_calibration_date(self, stack_index: int) -> str:
    """EEPROM request: Query the date of the last thermal calibration.

    Returns:
      Calibration date in ISO format 'YYYY-MM-DD'.
    """
    resp = await self.send_command("RCD", stack_index=stack_index)
    date = resp.strip()
    if not date or len(date) != 10 or date.count("-") != 2:
      raise RuntimeError(f"Unexpected RCD response: {resp!r}")
    return date

  # TODO: Command Placeholders

  async def request_calibration_low(self, sensor: int, format: int) -> float:
    raise NotImplementedError("RCL (Report Calibration Low) not implemented yet.")

  async def request_calibration_high(self, sensor: int, format: int) -> float:
    raise NotImplementedError("RCH (Report Calibration High) not implemented yet.")

  async def request_whole_calibration_data(self, key: str) -> bytes:
    raise NotImplementedError("RWC (Read Whole Calibration Data) not implemented yet.")

  async def request_proportionality_factor(self) -> int:
    raise NotImplementedError("RPF (Report Proportionality Factor) not implemented yet.")

  async def set_max_allowed_device_temperature(self, key: str, temperature: int) -> None:
    raise NotImplementedError("SMT (Set Max Allowed Device Temperature) not implemented yet.")

  async def set_pid_proportional_gain(self, key: str, value: int) -> None:
    raise NotImplementedError("SPP (Set PID Proportional Gain) not implemented yet.")

  async def set_pid_integration_value(self, key: str, value: int) -> None:
    raise NotImplementedError("SPI (Set PID Integration Value) not implemented yet.")

  async def delete_counter(self, key: str, selector: int) -> None:
    raise NotImplementedError("SDC (Set Delete Counter) not implemented yet.")

  async def set_boost_offset(self, offset: int) -> None:
    raise NotImplementedError("SBO (Set Boost Offset) not implemented yet.")

  async def set_boost_time(self, time_s: int) -> None:
    raise NotImplementedError("SBT (Set Boost Time) not implemented yet.")

  async def set_cooldown_time_factor(self, value: int) -> None:
    raise NotImplementedError("SHK (Set Cool-Down Time Evaluation Factor) not implemented yet.")

  async def set_heatup_time_factor(self, value: int) -> None:
    raise NotImplementedError("SHH (Set Heat-Up Time Evaluation Factor) not implemented yet.")

  async def set_heatup_offset(self, offset: int) -> None:
    raise NotImplementedError("SHO (Set Heat-Up Offset) not implemented yet.")

  async def set_calibration_low(self, key: str, sensor1: int, sensor2: int, sensor3: int) -> None:
    raise NotImplementedError("SCL (Set Calibration Low) not implemented yet.")

  async def set_calibration_high(
    self, key: str, sensor1: int, sensor2: int, sensor3: int, date: str
  ) -> None:
    raise NotImplementedError("SCH (Set Calibration High and Date) not implemented yet.")

  async def reset_calibration_data(self, key: str) -> None:
    raise NotImplementedError("SRC (Set Reset Calibration-Data) not implemented yet.")

  async def set_proportionality_factor(self, value: int) -> None:
    raise NotImplementedError("SPF (Set Proportionality Factor) not implemented yet.")

    # # # Setup Requirement # # #

  async def initialize(self, stack_index: int) -> str:
    """Initializes the machine unit after power-on.
    All other Action-Commands of the device are prohibited before the command is executed.
    On command AID, the drawer will be closed. If the drawer is hold in its open position,
    no Error will be generated!
    If AID is send during operating, the shaker and heater stop immediately!
    """
    resp = await self.send_command("AID", stack_index=stack_index)
    return resp

  # # # Loading Tray Features # # #

  async def open(self, stack_index: int) -> None:
    """Open the incubator door & move loading tray out."""
    await self.send_command("AOD", stack_index=stack_index)

  async def close(self, stack_index: int) -> None:
    """Move the loading tray in & close the incubator door."""
    await self.send_command("ACD", stack_index=stack_index)

  DrawerStatus = Literal["open", "closed"]

  async def request_drawer_status(self, stack_index: int) -> DrawerStatus:
    """Report the current drawer (loading tray) status.

    Returns:
      'open' if the loading tray is open, 'closed' if closed.

    Notes:
      - Firmware response: '1' = open, '0' = closed.
    """
    resp = await self.send_command("RDS", stack_index=stack_index)
    if resp == "1":
      return "open"
    if resp == "0":
      return "closed"
    raise ValueError(f"Unexpected RDS response: {resp!r}")

  # TODO: Drawer Placeholder Commands

  async def request_motor_power_clockwise(self) -> int:
    raise NotImplementedError("RPR (Report Motor Power Clockwise) not implemented yet.")

  async def request_motor_power_anticlockwise(self) -> int:
    raise NotImplementedError("RPL (Report Motor Power Anticlockwise) not implemented yet.")

  async def request_motor_current_limit_clockwise(self) -> int:
    raise NotImplementedError("RGR (Report Motor Current Limit Clockwise) not implemented yet.")

  async def request_motor_current_limit_anticlockwise(self) -> int:
    raise NotImplementedError("RGL (Report Motor Current Limit Anticlockwise) not implemented yet.")

  async def set_motor_power_clockwise(self, key: str, power: int) -> None:
    raise NotImplementedError("SPR (Set Motor Power Clockwise) not implemented yet.")

  async def set_motor_power_anticlockwise(self, key: str, power: int) -> None:
    raise NotImplementedError("SPL (Set Motor Power Anticlockwise) not implemented yet.")

  async def set_motor_current_limit_clockwise(self, key: str, current: int) -> None:
    raise NotImplementedError("SGR (Set Motor Current Limit Clockwise) not implemented yet.")

  async def set_motor_current_limit_anticlockwise(self, key: str, current: int) -> None:
    raise NotImplementedError("SGL (Set Motor Current Limit Anticlockwise) not implemented yet.")

  # # # Temperature Features # # #

  async def start_temperature_control(self, temperature: float, stack_index: int) -> None:
    """Set and activate the target incubation temperature (°C).

    The device begins active heating toward the target temperature.
    Passive cooling (firmware default) may occur automatically if the
    target temperature is below ambient, depending on environmental conditions.
    """

    assert temperature < self.max_temperature, (
      "Target temperature must be below max temperature of the incubator, i.e. "
      f"{self.max_temperature}C, target temperature given = {temperature}"
    )

    target = round(temperature * 10)
    await self.send_command(f"STT{target}", stack_index=stack_index)  # Store target temperature
    await self.send_command("SHE1", stack_index=stack_index)  # Enable temperature regulation

  async def stop_temperature_control(self, stack_index: int) -> None:
    """Stop active temperature regulation.

    Disables the incubator's heating control loop.
    The previously set target temperature remains stored in the
    device's memory but is no longer actively maintained.
    The incubator will passively drift toward ambient temperature.
    """
    await self.send_command("SHE0", stack_index=stack_index)

  async def get_temperature(
    self,
    stack_index: int,
    sensor: Literal["mean", "main", "dif", "boost"] = "main",
    read_timeout: float = 60.0,
  ) -> float:
    """Return current measured temperature in °C."""

    sensor_mapping = {
      "mean": [1, 2, 3],
      "main": [1],
      "dif": [2],
      "boost": [3],
    }
    vals = []
    for idx in sensor_mapping[sensor]:
      val = await self.send_command(f"RAT{idx}", stack_index=stack_index, read_timeout=read_timeout)
      vals.append(int(val) / 10.0)
    return round(sum(vals) / len(vals), 2)

  async def request_target_temperature(self, stack_index: int) -> float:
    """Return target temperature in °C."""

    resp = await self.send_command("RTT", stack_index=stack_index)

    return int(resp) / 10

  async def is_temperature_control_enabled(self, stack_index: int) -> bool:
    """
    Return True if active temperature control is enabled (RHE = 1 or 2),
    False if control is off (RHE = 0).

    Firmware response (RHE):
    0: control loop off (passive equilibrium)
    1: control loop on (heating/cooling active)
    2: control + booster on (extended heating mode)
    """
    resp = await self.send_command("RHE", stack_index=stack_index)
    try:
      status = int(resp.strip())
    except ValueError:
      raise InhecoError("RHE", "E00", f"Unexpected response: {resp!r}")

    if status not in (0, 1, 2):
      raise InhecoError("RHE", "E00", f"Invalid heater status value: {status}")

    enabled = status in (1, 2)

    return enabled

  async def request_pid_controller_coefficients(self, stack_index: int) -> tuple[float, float]:
    """
    Query the current PID controller coefficients.

    Returns:
        (P, I): tuple of floats
            - P: proportional gain (selector 1)
            - I: integration value (selector 2; 0 = integration off)
    """
    p_resp = await self.send_command("RPC1", stack_index=stack_index)
    i_resp = await self.send_command("RPC2", stack_index=stack_index)

    try:
      p = float(p_resp.strip())
      i = float(i_resp.strip())
    except ValueError:
      raise RuntimeError(f"Unexpected RPC response(s): P={p_resp!r}, I={i_resp!r}")

    return p, i

  async def request_maximum_allowed_temperature(
    self, stack_index: int, measured: bool = False
  ) -> float:
    """
    Query the maximum allowed or maximum measured device temperature (in °C).

    Args:
      measured:
      - False: report configured maximum allowed temperature (default)
      - True: report maximum measured temperature since last reset

    Returns:
      Temperature in °C (value / 10)
    """
    selector = "1" if measured else ""
    resp = await self.send_command(f"RMT{selector}", stack_index=stack_index)
    try:
      return int(resp.strip()) / 10.0
    except ValueError:
      raise RuntimeError(f"Unexpected RMT response: {resp!r}")

  async def request_delta_temperature(self, stack_index: int) -> float:
    """
    Query the absolute temperature difference between target and actual plate temperature.

    Returns:
      Delta temperature in °C (positive if below target, negative if above target).

    Notes:
    - Reported in 1/10 °C.
    - Negative values indicate the plate is warmer than the target.
    """
    resp = await self.send_command("RDT", stack_index=stack_index)
    try:
      return int(resp.strip()) / 10.0
    except ValueError:
      raise RuntimeError(f"Unexpected RDT response: {resp!r}")

  async def wait_for_temperature(
    self,
    stack_index: int,
    sensor: Literal["main", "dif", "boost", "mean"] = "main",
    tolerance: float = 0.2,
    interval_s: float = 0.5,
    timeout_s: Optional[float] = 600.0,
    show_progress_bar: bool = False,
  ) -> float:
    """
    Wait asynchronously until the target temperature is reached.

    Args:
      sensor: Temperature sensor to monitor ("main", "dif", "boost", or "mean").
      tolerance: Acceptable difference (in °C) between current and target temperature.
      interval_s: Polling interval in seconds. Default = 0.5 s.
      timeout_s: Maximum time to wait in seconds. None disables timeout. Default = 600 s.
      show_progress_bar: If True, display a dynamic ASCII progress bar in stdout. Default False.

    Returns:
      Final measured temperature in °C once within tolerance.

    Raises:
      TimeoutError: If target not reached within `timeout_s`.
      ValueError: If temperature control is not enabled or no valid target returned.
    """
    target_temp = await self.request_target_temperature(stack_index=stack_index)
    if target_temp is None:
      raise ValueError("Device did not return a valid target temperature.")

    temperature_control_enabled = await self.is_temperature_control_enabled(stack_index=stack_index)
    if not temperature_control_enabled:
      raise ValueError(
        f"Temperature control is not enabled on the machine ({stack_index}: {self.unit_composition[stack_index]})."
      )

    start_time = asyncio.get_event_loop().time()
    first_temp = await self.get_temperature(sensor=sensor, stack_index=stack_index)
    initial_diff = abs(first_temp - target_temp)
    bar_width = 40

    if show_progress_bar:
      print(f"Waiting for target temperature {target_temp:.2f} °C...\n")

    while True:
      current_temp = await self.get_temperature(sensor=sensor, stack_index=stack_index)
      diff = abs(current_temp - target_temp)

      # Compute normalized progress (1 = done)
      progress = 1.0 - min(diff / max(initial_diff, 1e-6), 1.0)
      filled = int(bar_width * progress)
      bar = "█" * filled + "-" * (bar_width - filled)

      if show_progress_bar:
        # Compute slope (°C/sec) based on direction of travel
        delta_done = abs(current_temp - first_temp)

        elapsed = asyncio.get_event_loop().time() - start_time

        slope = delta_done / max(elapsed, 1e-6)  # °C per second

        if slope > 0.0001:
          remaining = diff  # remaining °C
          eta_s = remaining / slope
          eta_str = f"{eta_s:6.1f}s"
        else:
          eta_str = "   --- "

        sys.stdout.write(f"\r[{bar}] {current_temp:.2f} °C  (Δ={diff:.2f} °C | ETA: {eta_str})")
        sys.stdout.flush()

      if diff <= tolerance:
        if show_progress_bar:
          sys.stdout.write("\n[OK] Target temperature reached.\n")
          sys.stdout.flush()

        self.logger.info("Target temperature reached (%.2f °C).", current_temp)
        return current_temp

      if timeout_s is not None:
        elapsed = asyncio.get_event_loop().time() - start_time
        if elapsed > timeout_s:
          if show_progress_bar:
            sys.stdout.write("\n[ERROR] Timeout waiting for temperature.\n")
            sys.stdout.flush()

          raise TimeoutError(
            f"Timeout after {timeout_s:.1f}s: "
            f"temperature {current_temp:.2f} °C "
            f"did not reach target {target_temp:.2f} °C ±{tolerance:.2f} °C."
          )

      await asyncio.sleep(interval_s)

  # # # Shaking Features # # #

  @requires_incubator_shaker
  async def request_shaker_frequency_x(self, stack_index: int, selector: int = 0) -> float:
    """Read the set or actual shaker frequency in the X-direction.

    Args:
      selector: 0 = to-be-set frequency, 1 = actual frequency.  Default = 0.

    Returns:
      Frequency in Hz.
    """
    if selector not in (0, 1):
      raise ValueError(f"Selector must be 0 or 1, got {selector}")
    resp = await self.send_command(f"RFX{selector}", stack_index=stack_index)
    return float(resp) / 10.0  # firmware reports in 1/10 Hz

  @requires_incubator_shaker
  async def request_shaker_frequency_y(self, stack_index: int, selector: int = 0) -> float:
    """Read the set or actual shaker frequency in the Y-direction.

    Args:
      selector: 0 = to-be-set frequency, 1 = actual frequency.  Default = 0.

    Returns:
      Frequency in Hz.
    """
    if selector not in (0, 1):
      raise ValueError(f"Selector must be 0 or 1, got {selector}")
    resp = await self.send_command(f"RFY{selector}", stack_index=stack_index)
    return float(resp) / 10.0  # firmware reports in 1/10 Hz

  @requires_incubator_shaker
  async def request_shaker_amplitude_x(self, stack_index: int, selector: int = 0) -> float:
    """Read the set, actual, or static shaker amplitude in the X-direction.

    Args:
      selector: 0 = set amplitude, 1 = actual amplitude, 2 = static distance from middle.  Default = 0.

    Returns:
      Amplitude in millimeters (mm).
    """
    if selector not in (0, 1, 2):
      raise ValueError(f"Selector must be 0, 1, or 2, got {selector}")
    resp = await self.send_command(f"RAX{selector}", stack_index=stack_index)
    return float(resp) / 10.0  # firmware reports in 1/10 mm

  @requires_incubator_shaker
  async def request_shaker_amplitude_y(self, stack_index: int, selector: int = 0) -> float:
    """Read the set, actual, or static shaker amplitude in the Y-direction.

    Args:
      selector: 0 = set amplitude, 1 = actual amplitude, 2 = static distance from middle.  Default = 0.

    Returns:
      Amplitude in millimeters (mm).
    """
    if selector not in (0, 1, 2):
      raise ValueError(f"Selector must be 0, 1, or 2, got {selector}")
    resp = await self.send_command(f"RAY{selector}", stack_index=stack_index)
    return float(resp) / 10.0  # firmware reports in 1/10 mm

  async def is_shaking_enabled(self, stack_index: int) -> bool:
    """Return True if the shaker is currently enabled or still decelerating.

    The firmware returns: 0 = shaker off; 1 = shaker on; 2 = shaker switched off but still moving.

    Returns:
      True if the shaker is active or still moving (status 1 or 2), False if fully stopped (status 0).
    """

    if "shaker" not in self.unit_composition[stack_index]:
      return False

    resp = await self.send_command("RSE", stack_index=stack_index)

    try:
      status = int(resp)
    except ValueError:
      raise InhecoError("RSE", "E00", f"Unexpected response: {resp!r}")

    if status not in (0, 1, 2):
      raise InhecoError("RSE", "E00", f"Invalid shaker status value: {status}")

    return status in (1, 2)  # TODO: discuss whether 2 should count as "shaking"

  @requires_incubator_shaker
  async def set_shaker_parameters(
    self,
    amplitude_x: float,
    amplitude_y: float,
    frequency_x: float,
    frequency_y: float,
    phase_shift: float,
    stack_index: int,
  ) -> None:
    """Set shaker parameters for both X and Y axes in a single command (firmware 'SSP').

    This combines the functionality of the individual SAX, SAY, SFX, SFY, and SPS commands.

    Args:
      amplitude_x: Amplitude on the X-axis in mm (0.0-3.0 mm, corresponds to 0-30 in firmware units).
      amplitude_y: Amplitude on the Y-axis in mm (0.0-3.0 mm, corresponds to 0-30 in firmware units).
      frequency_x: Frequency on the X-axis in Hz (6.6-30.0 Hz, corresponds to 66-300 in firmware units).
      frequency_y: Frequency on the Y-axis in Hz (6.6-30.0 Hz, corresponds to 66-300 in firmware units).
      phase_shift: Phase shift between X and Y axes in degrees (0-360°).

    Notes:
      - This command simplifies coordinated shaker setup.
      - All arguments are automatically converted to the firmware's expected integer scaling.
        (mm → x10; Hz → x10; ° left unscaled)
      - The firmware returns an acknowledgment frame on success.

    Raises:
      ValueError: If any parameter is outside its valid range.
      InhecoError: If the device reports an error or rejects the command.
    """
    # --- Validation ---
    if not (0.0 <= amplitude_x <= 3.0):
      raise ValueError(f"Amplitude X must be between 0.0 and 3.0 mm, got {amplitude_x}")
    if not (0.0 <= amplitude_y <= 3.0):
      raise ValueError(f"Amplitude Y must be between 0.0 and 3.0 mm, got {amplitude_y}")
    if not (6.6 <= frequency_x <= 30.0):
      raise ValueError(f"Frequency X must be between 6.6 and 30.0 Hz, got {frequency_x}")
    if not (6.6 <= frequency_y <= 30.0):
      raise ValueError(f"Frequency Y must be between 6.6 and 30.0 Hz, got {frequency_y}")
    if not (0.0 <= phase_shift <= 360.0):
      raise ValueError(f"Phase shift must be between 0° and 360°, got {phase_shift}")

    # --- Convert to firmware units ---
    amp_x_fw = round(amplitude_x * 10)
    amp_y_fw = round(amplitude_y * 10)
    freq_x_fw = round(frequency_x * 10)
    freq_y_fw = round(frequency_y * 10)
    phase_fw = round(phase_shift)

    # --- Build and send command ---
    cmd = f"SSP{amp_x_fw},{amp_y_fw},{freq_x_fw},{freq_y_fw},{phase_fw}"
    await self.send_command(cmd, stack_index=stack_index)

  def _mm_to_fw(self, mm: float) -> int:
    """Convert mm → firmware units (1/10 mm).

    Valid range: 0.0-3.0 mm (→ 0-30 in firmware).
    Raises ValueError if out of range.
    """
    if not (0.0 <= mm <= 3.0):
      raise ValueError(f"Amplitude must be between 0.0 and 3.0 mm, got {mm}")
    return int(round(mm * 10))

  def _rpm_to_fw_hz10(self, rpm: float) -> int:
    """Convert RPM → firmware Hz·10 units (validated).

    396-1800 RPM ↔ 6.6-30.0 Hz ↔ 66-300 in firmware.
    """
    if not (396 <= rpm <= 1800):
      raise ValueError(f"RPM must be between 396 and 1800, got {rpm}")
    return int(round((rpm / 60.0) * 10))

  def _hz_to_fw_hz10(self, hz: float) -> int:
    """Convert Hz → firmware Hz·10 units (validated)."""
    if not (6.6 <= hz <= 30.0):
      raise ValueError(f"Frequency must be between 6.6 and 30.0 Hz, got {hz}")
    return int(round(hz * 10))

  def _validate_hz_or_rpm(self, frequency_hz: Optional[float], rpm: Optional[float]) -> None:
    """Ensure exactly one of frequency_hz or rpm is provided."""
    if (frequency_hz is None) == (rpm is None):
      raise ValueError("Provide exactly one of frequency_hz or rpm.")

  def _phase_or_default(self, phase_deg: Optional[float], default: int) -> int:
    """Return integer phase or default (0-360°)."""
    p = default if phase_deg is None else int(round(phase_deg))
    if not (0 <= p <= 360):
      raise ValueError(f"Phase must be 0-360°, got {p}")
    return p

  def _fw_freq_pair(self, frequency_hz: Optional[float], rpm: Optional[float]) -> tuple[int, int]:
    """Return validated firmware frequency pair (Hz·10, Hz·10)."""
    if frequency_hz is not None:
      f = self._hz_to_fw_hz10(frequency_hz)
    else:
      # At this point rpm MUST be not None (validated earlier)
      assert rpm is not None
      f = self._rpm_to_fw_hz10(rpm)
    return (f, f)

  def _fw_amp_pair_linear_x(self, ax_mm: float) -> tuple[int, int]:
    return (self._mm_to_fw(ax_mm), 0)

  def _fw_amp_pair_linear_y(self, ay_mm: float) -> tuple[int, int]:
    return (0, self._mm_to_fw(ay_mm))

  def _fw_amp_pair_xy(self, ax_mm: float, ay_mm: float) -> tuple[int, int]:
    return (self._mm_to_fw(ax_mm), self._mm_to_fw(ay_mm))

  @requires_incubator_shaker
  async def set_shaker_pattern(
    self,
    *,
    pattern: Literal["linear_x", "linear_y", "orbital", "elliptical", "figure_eight"],
    stack_index: int,
    frequency_hz: Optional[float] = None,
    rpm: Optional[float] = None,
    amplitude_x_mm: Optional[float] = None,
    amplitude_y_mm: Optional[float] = None,
    phase_deg: Optional[float] = None,
  ) -> None:
    """Set the shaker motion pattern and parameters (without enabling motion).

    Patterns:
    - linear_x: motion along X only.
    - linear_y: motion along Y only.
    - orbital: circular motion (equal amplitudes on both axes, 90° phase).
    - elliptical: elliptical motion (unequal amplitudes, 90° phase).
    - figure_eight: double-loop motion (any amplitudes, 180° phase).
    """
    self._validate_hz_or_rpm(frequency_hz, rpm)
    fx, fy = self._fw_freq_pair(frequency_hz, rpm)

    if pattern == "linear_x":
      if amplitude_x_mm is None:
        raise ValueError("linear_x requires amplitude_x_mm.")
      ax, ay = self._fw_amp_pair_linear_x(amplitude_x_mm)
      phase = self._phase_or_default(phase_deg, 0)

    elif pattern == "linear_y":
      if amplitude_y_mm is None:
        raise ValueError("linear_y requires amplitude_y_mm.")
      ax, ay = self._fw_amp_pair_linear_y(amplitude_y_mm)
      phase = self._phase_or_default(phase_deg, 0)

    elif pattern == "orbital":
      # --- orbital: equal amplitudes, 90° phase ---
      if amplitude_x_mm is None or amplitude_y_mm is None:
        raise ValueError("orbital requires both amplitude_x_mm and amplitude_y_mm.")
      if abs(amplitude_x_mm - amplitude_y_mm) > 1e-6:
        raise ValueError(
          f"Orbital motion requires equal amplitudes on X and Y "
          f"(got {amplitude_x_mm} mm vs {amplitude_y_mm} mm). "
          f"Use pattern='elliptical' instead."
        )
      ax, ay = self._fw_amp_pair_xy(amplitude_x_mm, amplitude_y_mm)
      phase = self._phase_or_default(phase_deg, 90)

    elif pattern == "elliptical":
      # --- elliptical: differing amplitudes, 90° phase ---
      ax_mm = amplitude_x_mm if amplitude_x_mm is not None else 2.5
      ay_mm = amplitude_y_mm if amplitude_y_mm is not None else 2.0
      ax, ay = self._fw_amp_pair_xy(ax_mm, ay_mm)
      phase = self._phase_or_default(phase_deg, 90)

    elif pattern == "figure_eight":
      # --- true figure eight: fx:fy = 1:2, phase = 90° ---
      ax_mm = amplitude_x_mm if amplitude_x_mm is not None else 2.5
      ay_mm = amplitude_y_mm if amplitude_y_mm is not None else 2.5
      ax, ay = self._fw_amp_pair_xy(ax_mm, ay_mm)

      # base frequency (default 10 Hz if not given)
      base_hz = frequency_hz if frequency_hz is not None else (rpm / 60.0 if rpm else 10.0)
      fx = self._hz_to_fw_hz10(base_hz)
      fy = self._hz_to_fw_hz10(base_hz * 2)

      phase = self._phase_or_default(phase_deg, 90)

    else:
      raise ValueError(
        f"Unknown pattern: {pattern}"
        "\nValid options: 'linear_x', 'linear_y', 'orbital', 'elliptical', 'figure_eight'"
      )

    await self.send_command(f"SSP{ax},{ay},{fx},{fy},{phase}", stack_index=stack_index)

  @requires_incubator_shaker
  async def set_shaker_status(self, enabled: bool, stack_index: int) -> None:
    """Enable or disable the shaker (ASEND always used when enabled)."""
    await self.send_command("ASEND" if enabled else "ASE0", stack_index=stack_index)

  @requires_incubator_shaker
  async def shake(
    self,
    *,
    stack_index: int,
    pattern: Literal["linear_x", "linear_y", "orbital", "elliptical", "figure_eight"] = "orbital",
    rpm: Optional[float] = None,
    frequency_hz: Optional[float] = None,
    amplitude_x_mm: float = 3.0,
    amplitude_y_mm: float = 3.0,
    phase_deg: Optional[float] = None,
  ) -> None:
    """
    Configure and start shaking with the given motion pattern.

    This command safely updates shaker parameters (frequency, amplitude, phase)
    and starts motion using `ASEND` (no labware detection). If the shaker is
    already running, it is first stopped and reinitialized before applying new
    parameters—required because the firmware only latches `SSP` settings when
    the shaker transitions from idle to active.

    Behavior:
    - Stops the shaker if active, waits briefly, applies the new pattern, and restarts shaking.
    - Ensures consistent parameter changes and prevents ignored SSP updates.

    Args:
      pattern: Motion pattern: `"linear_x"`, `"linear_y"`, `"orbital"`, `"elliptical"`, or `"figure_eight"`.
      rpm: Rotational speed (396-1800 RPM). Mutually exclusive with `frequency_hz`.
      frequency_hz: Oscillation frequency (6.6-30.0 Hz). Mutually exclusive with `rpm`.
      amplitude_x_mm: X-axis amplitude in mm (0.0-3.0 mm).
      amplitude_y_mm: Y-axis amplitude in mm (0.0-3.0 mm).
      phase_deg: Optional phase offset between X and Y axes (0-360°).

    Raises:
      ValueError: If parameter ranges or combinations are invalid.
      InhecoError: If the device rejects the command or is not ready.
    """

    is_shaking = await self.is_shaking_enabled(stack_index=stack_index)
    if is_shaking:
      await self.stop_shaking(stack_index=stack_index)
      await asyncio.sleep(0.5)  # brief pause for firmware to settle

    await self.set_shaker_pattern(
      pattern=pattern,
      rpm=rpm,
      stack_index=stack_index,
      frequency_hz=frequency_hz,
      amplitude_x_mm=amplitude_x_mm,
      amplitude_y_mm=amplitude_y_mm,
      phase_deg=phase_deg,
    )
    await self.set_shaker_status(True, stack_index=stack_index)

  # TODO: expose direction argument -> clockwise / counterclockwise for rotating shaking patterns

  @requires_incubator_shaker
  async def stop_shaking(self, stack_index: int) -> None:
    """Stop shaker (ASE0)."""

    await self.set_shaker_status(False, stack_index=stack_index)

  @requires_incubator_shaker
  async def request_shaker_phase_shift(self, stack_index: int, selector: int = 0) -> float:
    """Read the set or actual phase shift between X and Y shaker drives (firmware 'RPS' command).

    Args:
      selector: 0 = currently set phase shift, 1 = actual phase shift.  Default = 0.

    Returns:
      Phase shift in degrees [°].  Returns 12345.0 if the shaker has not reached a stable state or if phase shift calculation is invalid due to too-small amplitudes (< 1 mm on either axis).
    """
    if selector not in (0, 1):
      raise ValueError(f"Selector must be 0 or 1, got {selector}")

    resp = await self.send_command(f"RPS{selector}", stack_index=stack_index)
    return float(resp)

  @requires_incubator_shaker
  async def request_shaker_calibration_value(self, position: int, selector: int = 0) -> float:
    raise NotImplementedError("RSC (Read Shaker Calibration Values) not implemented yet.")

  @requires_incubator_shaker
  async def read_whole_shaker_calibration_data(self, key: str) -> str:
    raise NotImplementedError("RWJ (Read Whole Shaker Adjustment Data) not implemented yet.")

  @requires_incubator_shaker
  async def set_shaker_calibration_value(
    self,
    key: str,
    position: int,
    value: int,
  ) -> None:
    raise NotImplementedError("SSC (Set Shaker Calibration Values) not implemented yet.")

  # # # Self-Test # # #

  async def perform_self_test(self, stack_index: int, read_timeout: int = 500) -> Dict[str, bool]:
    """Execute the internal self-test routine.

    Normal Testing-Time: ca. 3 minutes (Beware of timeouts!)
    Maximum Testing-Time: 465 sec. (Beware of timeouts!)
    The Test must be performed without a Labware inside!
    The Error code is reported as an 11-Bit-Code. A set bit (1) represents an Error. The Test stops
    immediately, when the Drawer-Test fails to avoid damage to the device!
    Please note the AQS was developed for the Incubator Shaker MP. This can lead assemblies to
    DWP Incubators that the result of the AQS, due to the design, 8 (bit2) could be.

    Returns:
      A dictionary mapping error condition names to booleans indicating presence of each error.
    """

    plate_in_status = await self.request_plate_in_incubator(stack_index=stack_index)
    if plate_in_status:
      raise ValueError("Self-test requires an empty incubator.")

    loading_tray_status = await self.request_drawer_status(stack_index=stack_index)
    if loading_tray_status == "open":
      raise ValueError("Self-test requires a closed loading tray.")

    resp = await self.send_command("AQS", stack_index=stack_index, read_timeout=read_timeout)

    if not isinstance(resp, str):
      raise RuntimeError(f"Invalid response: {resp!r}")

    resp_decimal = int(resp)

    assert 0 <= resp_decimal <= 2047, f"Invalid self-test response received: {resp!r}"

    binary_response = bin(resp_decimal)[2:].zfill(11)
    return {
      "drawer_error": binary_response[-1] == "1",
      "homogeneity_sensor_3_vs_1_error": binary_response[-2] == "1",
      "homogeneity_sensor_2_vs_1_error": binary_response[-3] == "1",
      "sensor_1_target_temp_error": binary_response[-4] == "1",
      "y_amplitude_shaker_error": binary_response[-5] == "1",
      "x_amplitude_shaker_error": binary_response[-6] == "1",
      "phase_shift_shaker_error": binary_response[-7] == "1",
      "y_frequency_shaker_error": binary_response[-8] == "1",
      "x_frequency_shaker_error": binary_response[-9] == "1",
      "line_boost_heater_broken": binary_response[-10] == "1",
      "line_main_heater_broken": binary_response[-11] == "1",
    }


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# IncubatorShakerUnit Class
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class InhecoIncubatorShakerUnit:
  """High-level API for an individual Inheco Incubator/Shaker unit in a stacked system."""

  def __init__(self, backend: InhecoIncubatorShakerStackBackend, index: int):
    self.backend: InhecoIncubatorShakerStackBackend = backend
    self.index = index

  def __repr__(self):
    return f"<InhecoIncubatorShakerUnit index={self.index}>"

  # === Common high-level API shortcuts (explicitly delegate) ===

  # Querying Machine State #
  async def request_firmware_version(self) -> str:
    """EEPROM request: Return the firmware version string."""
    return await self.backend.request_firmware_version(stack_index=self.index)

  async def request_serial_number(self) -> str:
    """EEPROM request: Return the device serial number."""
    return await self.backend.request_serial_number(stack_index=self.index)

  async def request_last_calibration_date(self) -> str:
    """EEPROM request: Query the date of the last calibration."""
    return await self.backend.request_last_calibration_date(stack_index=self.index)

  async def request_machine_allocation(self, layer: int = 0) -> dict:
    """
    Report which device slots are occupied on a given layer (firmware 'RDAx,0').

    Args:
      layer: Layer index (0-7). Default = 0.

    Returns:
      {
        "layer": int,
        "slot_mask": int,       # e.g. 7
        "slot_mask_bin": str,   # e.g. "0b0000000000000111"
        "slots_connected": list[int]  # e.g. [0, 1, 2]
      }

    Notes:
      Each bit in `slot_mask` represents one of 16 possible device slots: bit=1 means a device is connected; bit=0 means empty.
    """
    return await self.backend.request_machine_allocation(layer=layer, stack_index=self.index)

  async def request_number_of_connected_machines(self, layer: int = 0) -> int:
    """
    Report the number of connected Inheco devices on a layer (RDAx,1).

    Args:
      layer: Layer index (0-7). Default = 0.

    Returns:
      Number of connected devices (0-16).

    Example:
      Response "3" → 3 connected devices on that layer.
    """
    return await self.backend.request_number_of_connected_machines(
      layer=layer, stack_index=self.index
    )

  async def request_labware_detection_threshold(self) -> int:
    """EEPROM request"""
    return await self.backend.request_labware_detection_threshold(stack_index=self.index)

  async def request_incubator_type(self) -> InhecoIncubatorUnitType:
    """Return a descriptive string of the incubator/shaker configuration."""
    return await self.backend.request_incubator_type(stack_index=self.index)

  async def request_plate_in_incubator(self) -> bool:
    """Sensor command: Check if a plate is currently present in the incubator."""
    return await self.backend.request_plate_in_incubator(stack_index=self.index)

  async def request_operation_time_in_hours(self) -> int:
    """EEPROM request: Total operation time of the device in hours."""
    return await self.backend.request_operation_time_in_hours(stack_index=self.index)

  async def request_drawer_cycles_performed(self) -> int:
    """EEPROM request: Number of drawer open/close cycles performed."""
    return await self.backend.request_drawer_cycles_performed(stack_index=self.index)

  async def request_is_initialized(self) -> bool:
    """EEPROM request: Check if the machine has been initialized."""
    resp = await self.backend.request_is_initialized(stack_index=self.index)
    return resp in {"0", "2"}

  async def request_plate_status_known(self) -> bool:
    """EEPROM request: Check if the plate status is known."""
    return await self.backend.request_plate_status_known(stack_index=self.index)

  async def request_thermal_calibration_date(self) -> str:
    """EEPROM request: Query the date of the last thermal calibration.

    Returns:
      Calibration date in ISO format 'YYYY-MM-DD'.
    """
    return await self.backend.request_thermal_calibration_date(stack_index=self.index)

  # # # Setup Requirement # # #

  async def initialize(self) -> str:
    """Initializes the machine unit after power-on.
    All other Action-Commands of the device are prohibited before the command is executed.
    On command AID, the drawer will be closed. If the drawer is hold in its open position,
    no Error will be generated!
    If AID is send during operating, the shaker and heater stop immediately!
    """
    return await self.backend.initialize(stack_index=self.index)

  # # # Loading Tray Features # # #

  async def open(self) -> None:
    """Open the incubator door & move loading tray out."""
    await self.backend.open(stack_index=self.index)

  async def close(self) -> None:
    """Move the loading tray in & close the incubator door."""
    await self.backend.close(stack_index=self.index)

  async def request_drawer_status(self) -> InhecoIncubatorShakerStackBackend.DrawerStatus:
    """Report the current drawer (loading tray) status.

    Returns:
      'open' if the loading tray is open, 'closed' if closed.
    """
    return await self.backend.request_drawer_status(stack_index=self.index)

  # # # Temperature Features # # #

  async def start_temperature_control(self, temperature: float) -> None:
    """Set and activate the target incubation temperature (°C).

    The device begins active heating toward the target temperature.
    Passive cooling (firmware default) may occur automatically if the
    target temperature is below ambient, depending on environmental conditions.
    """
    await self.backend.start_temperature_control(temperature=temperature, stack_index=self.index)

  async def stop_temperature_control(self) -> None:
    """Stop active temperature regulation.

    Disables the incubator's heating control loop.
    The previously set target temperature remains stored in the
    device's memory but is no longer actively maintained.
    The incubator will passively drift toward ambient temperature.
    """
    await self.backend.stop_temperature_control(stack_index=self.index)

  async def get_temperature(
    self,
    sensor: Literal["mean", "main", "dif", "boost"] = "main",
    read_timeout: float = 60.0,
  ) -> float:
    """Return current measured temperature in °C."""
    return await self.backend.get_temperature(
      sensor=sensor, stack_index=self.index, read_timeout=read_timeout
    )

  async def request_target_temperature(self) -> float:
    """Return target temperature in °C."""
    return await self.backend.request_target_temperature(stack_index=self.index)

  async def is_temperature_control_enabled(self) -> bool:
    """
    Return True if active temperature control is enabled (RHE = 1 or 2),
    False if control is off (RHE = 0).

    Firmware response (RHE): 0: control loop off (passive equilibrium).  1: control loop on (heating/cooling active).  2: control + booster on (extended heating mode).
    """
    return await self.backend.is_temperature_control_enabled(stack_index=self.index)

  async def request_pid_controller_coefficients(self) -> tuple[float, float]:
    """
    Query the current PID controller coefficients.

    Returns:
      (P, I): tuple of floats. P: proportional gain (selector 1). I: integration value (selector 2; 0 = integration off)
    """
    return await self.backend.request_pid_controller_coefficients(stack_index=self.index)

  async def request_maximum_allowed_temperature(self, measured: bool = False) -> float:
    """
    Query the maximum allowed or maximum measured device temperature (in °C).

    Args:
      measured:
        - False: report configured maximum allowed temperature (default)
        - True: report maximum measured temperature since last reset

    Returns:
      Temperature in °C (value / 10)
    """
    return await self.backend.request_maximum_allowed_temperature(
      stack_index=self.index, measured=measured
    )

  async def request_delta_temperature(self) -> float:
    """
    Query the absolute temperature difference between target and actual plate temperature.

    Returns:
      Delta temperature in °C (positive if below target, negative if above target).

    Notes:
    - Reported in 1/10 °C.
    - Negative values indicate the plate is warmer than the target.
    """
    return await self.backend.request_delta_temperature(stack_index=self.index)

  async def wait_for_temperature(
    self,
    *,
    sensor: Literal["main", "dif", "boost", "mean"] = "main",
    tolerance: float = 0.2,
    interval_s: float = 0.5,
    timeout_s: Optional[float] = 600.0,
    show_progress_bar: bool = False,
  ) -> float:
    """
    Wait asynchronously until the target temperature is reached.

    Args:
      sensor: Temperature sensor to monitor ("main", "dif", "boost", or "mean").
      tolerance: Acceptable difference (in °C) between current and target temperature.
      interval_s: Polling interval in seconds. Default = 0.5 s.
      timeout_s: Maximum time to wait in seconds. None disables timeout. Default = 600 s.
      show_progress_bar: If True, display a dynamic ASCII progress bar in stdout. Default False.

    Returns:
      Final measured temperature in °C once within tolerance.

    Raises:
      TimeoutError: If target not reached within `timeout_s`.
      ValueError: If temperature control is not enabled or no valid target returned.
    """
    return await self.backend.wait_for_temperature(
      stack_index=self.index,
      sensor=sensor,
      tolerance=tolerance,
      interval_s=interval_s,
      timeout_s=timeout_s,
      show_progress_bar=show_progress_bar,
    )

  # # # Shaking Features # # #

  async def request_shaker_frequency_x(self, selector: int = 0) -> float:
    """Read the set or actual shaker frequency in the X-direction.

    Args:
      selector: 0 = to-be-set frequency, 1 = actual frequency.  Default = 0.

    Returns:
      Frequency in Hz.
    """
    return await self.backend.request_shaker_frequency_x(stack_index=self.index, selector=selector)

  async def request_shaker_frequency_y(self, selector: int = 0) -> float:
    """Read the set or actual shaker frequency in the Y-direction.

    Args:
      selector: 0 = to-be-set frequency, 1 = actual frequency.  Default = 0.

    Returns:
      Frequency in Hz.
    """
    return await self.backend.request_shaker_frequency_y(stack_index=self.index, selector=selector)

  async def request_shaker_amplitude_x(self, selector: int = 0) -> float:
    """Read the set, actual, or static shaker amplitude in the X-direction.

    Args:
      selector: 0 = set amplitude, 1 = actual amplitude, 2 = static distance from middle.  Default = 0.

    Returns:
      Amplitude in millimeters (mm).
    """
    return await self.backend.request_shaker_amplitude_x(stack_index=self.index, selector=selector)

  async def request_shaker_amplitude_y(self, selector: int = 0) -> float:
    """Read the set, actual, or static shaker amplitude in the Y-direction.

    Args:
      selector: 0 = set amplitude, 1 = actual amplitude, 2 = static distance from middle.  Default = 0.

    Returns:
        Amplitude in millimeters (mm).
    """
    return await self.backend.request_shaker_amplitude_y(stack_index=self.index, selector=selector)

  async def is_shaking_enabled(self) -> bool:
    """Return True if the shaker is currently enabled or still decelerating.

    The firmware returns: 0: shaker off, 1: shaker on, 2: shaker switched off but still moving.

    Returns:
      True if the shaker is active or still moving (status 1 or 2), False if fully stopped (status 0).
    """
    return await self.backend.is_shaking_enabled(stack_index=self.index)

  async def set_shaker_parameters(
    self,
    amplitude_x: float,
    amplitude_y: float,
    frequency_x: float,
    frequency_y: float,
    phase_shift: float,
    stack_index: int,
  ) -> None:
    """Set shaker parameters for both X and Y axes in a single command (firmware 'SSP').

    This combines the functionality of the individual SAX, SAY, SFX, SFY, and SPS commands.

    Args:
      amplitude_x: Amplitude on the X-axis in mm (0.0-3.0 mm, corresponds to 0-30 in firmware units).
      amplitude_y: Amplitude on the Y-axis in mm (0.0-3.0 mm, corresponds to 0-30 in firmware units).
      frequency_x: Frequency on the X-axis in Hz (6.6-30.0 Hz, corresponds to 66-300 in firmware units).
      frequency_y: Frequency on the Y-axis in Hz (6.6-30.0 Hz, corresponds to 66-300 in firmware units).
      phase_shift: Phase shift between X and Y axes in degrees (0-360°).

    Notes:
    - This command simplifies coordinated shaker setup.
    - All arguments are automatically converted to the firmware's expected integer scaling.
      (mm → x10; Hz → x10; ° left unscaled)
    - The firmware returns an acknowledgment frame on success.

    Raises:
      ValueError: If any parameter is outside its valid range.
      InhecoError: If the device reports an error or rejects the command.
    """
    await self.backend.set_shaker_parameters(
      stack_index=stack_index,
      amplitude_x=amplitude_x,
      amplitude_y=amplitude_y,
      frequency_x=frequency_x,
      frequency_y=frequency_y,
      phase_shift=phase_shift,
    )

  async def set_shaker_pattern(
    self,
    *,
    pattern: Literal["linear_x", "linear_y", "orbital", "elliptical", "figure_eight"],
    frequency_hz: Optional[float] = None,
    rpm: Optional[float] = None,
    amplitude_x_mm: Optional[float] = None,
    amplitude_y_mm: Optional[float] = None,
    phase_deg: Optional[float] = None,
  ) -> None:
    """Set the shaker motion pattern and parameters (without enabling motion).

    Patterns:
      - linear_x:    motion along X only.
      - linear_y:    motion along Y only.
      - orbital:     circular motion (equal amplitudes on both axes, 90° phase).
      - elliptical:  elliptical motion (unequal amplitudes, 90° phase).
      - figure_eight: double-loop motion (any amplitudes, 180° phase).
    """
    await self.backend.set_shaker_pattern(
      stack_index=self.index,
      pattern=pattern,
      frequency_hz=frequency_hz,
      rpm=rpm,
      amplitude_x_mm=amplitude_x_mm,
      amplitude_y_mm=amplitude_y_mm,
      phase_deg=phase_deg,
    )

  async def set_shaker_status(self, enabled: bool) -> None:
    """Enable or disable the shaker (ASEND always used when enabled)."""
    await self.backend.set_shaker_status(stack_index=self.index, enabled=enabled)

  async def shake(
    self,
    *,
    pattern: Literal["linear_x", "linear_y", "orbital", "elliptical", "figure_eight"] = "orbital",
    rpm: Optional[float] = None,
    frequency_hz: Optional[float] = None,
    amplitude_x_mm: float = 3.0,
    amplitude_y_mm: float = 3.0,
    phase_deg: Optional[float] = None,
  ) -> None:
    """
    Configure and start shaking with the given motion pattern.

    This command safely updates shaker parameters (frequency, amplitude, phase)
    and starts motion using `ASEND` (no labware detection). If the shaker is
    already running, it is first stopped and reinitialized before applying new
    parameters—required because the firmware only latches `SSP` settings when
    the shaker transitions from idle to active.

    Behavior:
      - Stops the shaker if active, waits briefly, applies the new pattern, and restarts shaking.
      - Ensures consistent parameter changes and prevents ignored SSP updates.

    Args:
      pattern: Motion pattern: `"linear_x"`, `"linear_y"`, `"orbital"`, `"elliptical"`, or `"figure_eight"`.
      rpm: Rotational speed (396-1800 RPM). Mutually exclusive with `frequency_hz`.
      frequency_hz: Oscillation frequency (6.6-30.0 Hz). Mutually exclusive with `rpm`.
      amplitude_x_mm: X-axis amplitude in mm (0.0-3.0 mm).
      amplitude_y_mm: Y-axis amplitude in mm (0.0-3.0 mm).
      phase_deg: Optional phase offset between X and Y axes (0-360°).

    Raises:
      ValueError: If parameter ranges or combinations are invalid.
      InhecoError: If the device rejects the command or is not ready.
    """
    await self.backend.shake(
      stack_index=self.index,
      pattern=pattern,
      rpm=rpm,
      frequency_hz=frequency_hz,
      amplitude_x_mm=amplitude_x_mm,
      amplitude_y_mm=amplitude_y_mm,
      phase_deg=phase_deg,
    )

  async def stop_shaking(self) -> None:
    """Stop shaker (ASE0)."""
    await self.backend.stop_shaking(stack_index=self.index)

  async def request_shaker_phase_shift(self, selector: int = 0) -> float:
    """Read the set or actual phase shift between X and Y shaker drives (firmware 'RPS' command).

    Args:
      selector: 0 = currently set phase shift, 1 = actual phase shift.  Default = 0.

    Returns:
      Phase shift in degrees [°].  Returns 12345.0 if the shaker has not reached a stable state or if phase shift calculation is invalid due to too-small amplitudes (< 1 mm on either axis).
    """
    return await self.backend.request_shaker_phase_shift(stack_index=self.index, selector=selector)

  # # # Self-Test # # #

  async def perform_self_test(self, read_timeout: int = 500) -> Dict[str, bool]:
    """Execute the internal self-test routine."""

    return await self.backend.perform_self_test(stack_index=self.index, read_timeout=read_timeout)



from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.tilting.hamilton_backend import (
  HamiltonTiltModuleBackend,
)
from pylabrobot.tilting.tilter import Tilter


class HamiltonTiltModule(Tilter):
  """A Hamilton tilt module."""

  def __init__(
    self,
    name: str,
    com_port: str,
    child_location: Coordinate = Coordinate(1.0, 3.0, 83.55),
    pedestal_size_z: float = 3.47,
    write_timeout: float = 3,
    timeout: float = 3,
  ):
    """Initialize a Hamilton tilt module.

    Args:
      com_port: The communication port.
      child_location: The location of the child resource.
      pedestal_size_z: The size of the pedestal in the z dimension.
      write_timeout: The write timeout. Defaults to 3.
      timeout: The timeout. Defaults to 3.
    """

    super().__init__(
      name=name,
      size_x=132,
      size_y=92.57,
      size_z=85.81,
      backend=HamiltonTiltModuleBackend(
        com_port=com_port,
        write_timeout=write_timeout,
        timeout=timeout,
      ),
      hinge_coordinate=Coordinate(6.18, 0, 72.85),
      child_location=child_location,
      category="tilter",
      model=HamiltonTiltModule.__name__,
    )

    self.pedestal_size_z = pedestal_size_z



import textwrap
import unittest

from pylabrobot.resources.corning import (
  Cor_96_wellplate_360ul_Fb,
)
from pylabrobot.resources.hamilton import (
  PLT_CAR_L5AC_A00,
  TIP_CAR_480_A00,
  STARLetDeck,
  hamilton_96_tiprack_300uL_filter,
  hamilton_96_tiprack_1000uL_filter,
)
from pylabrobot.resources.stanley.cups import (
  StanleyCup_QUENCHER_FLOWSTATE_TUMBLER,
)


class HamiltonDeckTests(unittest.TestCase):
  """Tests for the HamiltonDeck class."""

  def build_layout(self):
    """Build a deck layout for testing"""
    deck = STARLetDeck()

    tip_car = TIP_CAR_480_A00(name="tip_carrier")
    tip_car[0] = hamilton_96_tiprack_300uL_filter(name="tip_rack_01")
    tip_car[1] = hamilton_96_tiprack_300uL_filter(name="tip_rack_02")
    tip_car[3] = hamilton_96_tiprack_1000uL_filter(name="tip_rack_04")

    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = Cor_96_wellplate_360ul_Fb(name="aspiration plate")
    plt_car[2] = Cor_96_wellplate_360ul_Fb(name="dispense plate")

    deck.assign_child_resource(tip_car, rails=1)
    deck.assign_child_resource(plt_car, rails=21)

    return deck

  def test_summary(self):
    self.maxDiff = None
    deck = self.build_layout()
    self.assertEqual(
      deck.summary(),
      textwrap.dedent(
        """
    Rail  Resource                      Type                 Coordinates (mm)
    =======================================================================================
    (-6)  ├── trash_core96              Trash                (-58.200, 106.000, 216.400)
          │
    (1)   ├── tip_carrier               TipCarrier           (100.000, 063.000, 100.000)
          │   ├── tip_rack_01           TipRack              (106.200, 073.000, 214.950)
          │   ├── tip_rack_02           TipRack              (106.200, 169.000, 214.950)
          │   ├── <empty>
          │   ├── tip_rack_04           TipRack              (106.200, 361.000, 214.950)
          │   ├── <empty>
          │
    (21)  ├── plate carrier             PlateCarrier         (550.000, 063.000, 100.000)
          │   ├── aspiration plate      Plate                (554.000, 071.500, 183.120)
          │   ├── <empty>
          │   ├── dispense plate        Plate                (554.000, 263.500, 183.120)
          │   ├── <empty>
          │   ├── <empty>
          │
    (31)  ├── waste_block               Resource             (775.000, 115.000, 100.000)
          │   ├── teaching_tip_rack     TipRack              (780.900, 461.100, 100.000)
          │   ├── core_grippers         HamiltonCoreGrippers (797.500, 085.500, 205.000)
          │
    (32)  ├── trash                     Trash                (800.000, 190.600, 137.100)
    """[1:]
      ),
    )

  def test_assign_gigantic_resource(self):
    stanley_cup = StanleyCup_QUENCHER_FLOWSTATE_TUMBLER(name="HUGE")
    deck = STARLetDeck()
    with self.assertLogs("pylabrobot") as log:
      deck.assign_child_resource(stanley_cup, rails=1)
    self.assertEqual(
      log.output,
      [
        "WARNING:pylabrobot:Resource 'HUGE' is very high on the deck: 412.42 mm. Be "
        "careful when traversing the deck.",
        "WARNING:pylabrobot:Resource 'HUGE' is very high on the deck: 412.42 mm. Be "
        "careful when grabbing this resource.",
      ],
    )



"""Module defining tube containers."""

import enum
import warnings
from typing import Callable, Dict, List, Optional, Tuple, Union

from pylabrobot.resources.container import Container
from pylabrobot.resources.liquid import Liquid


class TubeBottomType(enum.Enum):
  """Enum for the type of bottom of a tube."""

  FLAT = "flat"
  U = "U"
  V = "V"
  UNKNOWN = "unknown"


class Tube(Container):
  """Tube container, like Eppendorf tubes.

  Note that in regular use these will be automatically generated by the
  :class:`pylabrobot.resources.TubeRack` class.
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    max_volume: float,
    material_z_thickness: Optional[float] = None,
    category: str = "tube",
    model: Optional[str] = None,
    bottom_type: Union[TubeBottomType, str] = TubeBottomType.UNKNOWN,
    compute_volume_from_height: Optional[Callable[[float], float]] = None,
    compute_height_from_volume: Optional[Callable[[float], float]] = None,
    height_volume_data: Optional[Dict[float, float]] = None,
  ):
    """Create a new tube.

    Args:
      name: Name of the tube.
      size_x: Size of the tube in the x direction.
      size_y: Size of the tube in the y direction.
      size_z: Size of the tube in the z direction.
      material_z_thickness: Tube base to cavity base.
      max_volume: Maximum volume of the tube.
      category: Category of the tube.
      height_volume_data: Optional dict mapping height (mm) to volume (uL). See
        :class:`Container` for details.
    """
    if isinstance(bottom_type, str):
      bottom_type = TubeBottomType(bottom_type)

    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      material_z_thickness=material_z_thickness,
      category=category,
      max_volume=max_volume,
      model=model,
      compute_volume_from_height=compute_volume_from_height,
      compute_height_from_volume=compute_height_from_volume,
      height_volume_data=height_volume_data,
    )
    self.tracker.register_callback(self._state_updated)

  def serialize(self) -> dict:
    return {**super().serialize(), "max_volume": self.max_volume}

  def set_volume(self, volume: float):
    """Set the volume in the tube.

    (wraps :meth:`~.VolumeTracker.set_volume`)

    Example:
      Set the volume in a tube to 10 uL:

      >>> tube.set_volume(10)
    """

    self.tracker.set_volume(volume)

  def set_liquids(self, liquids: List[Tuple[Optional["Liquid"], float]]):
    """Set the liquids in the tube.

    Deprecated: use `set_volume` instead. This method will be removed in a future version.
    """

    warnings.warn(
      "`set_liquids` is deprecated and will be removed in a future version. Use `set_volume` instead.",
      DeprecationWarning,
      stacklevel=2,
    )

    self.tracker.set_liquids(liquids)



from __future__ import annotations

import logging
from typing import Dict, Generic, List, Optional, Type, TypeVar, Union, cast

from pylabrobot.resources.resource_holder import ResourceHolder, get_child_location

from .coordinate import Coordinate
from .plate import Plate
from .resource import Resource
from .resource_stack import ResourceStack

logger = logging.getLogger("pylabrobot")


S = TypeVar("S", bound=ResourceHolder)


class Carrier(Resource, Generic[S]):
  """Base class for all carriers."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    sites: Optional[Dict[int, S]] = None,
    category: Optional[str] = "carrier",
    model: Optional[str] = None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      category=category,
      model=model,
    )

    sites = sites or {}

    self.sites: Dict[int, S] = {}
    for spot, site in sites.items():
      if site.location is None:
        raise ValueError(f"site {site} has no location")
      self.assign_child_resource(site, location=site.location + get_child_location(site), spot=spot)

  @property
  def capacity(self):
    """The number of sites on this carrier."""
    return len(self.sites)

  def __len__(self) -> int:
    """Return the number of sites on this carrier."""
    return len(self.sites)

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate],
    reassign: bool = True,
    spot: Optional[int] = None,
  ):
    if not isinstance(resource, (ResourceHolder, Carrier)):
      raise TypeError(f"Invalid resource {resource}")

    # see if we have an index for the resource name (eg from deserialization or user specification),
    # otherwise add in first available spot
    idx = spot if spot is not None else len(self.sites)
    if not reassign and self.sites[idx] is not None:
      raise ValueError(f"a site with index {idx} already exists")
    self.sites[idx] = cast(S, resource)

    super().assign_child_resource(resource, location=location, reassign=reassign)

  def assign_resource_to_site(self, resource: Resource, spot: int):
    if self.sites[spot].resource is not None:
      raise ValueError(f"spot {spot} already has a resource")
    self.sites[spot].assign_child_resource(resource)

  def unassign_child_resource(self, resource: Resource):
    """Unassign a resource from this carrier, checked by name.

    Raises:
      ValueError: If the resource is not assigned to this carrier.
    """

    if not isinstance(resource.parent, ResourceHolder) or not resource.parent.parent == self:
      raise ValueError(f"Resource {resource} is not assigned to this carrier")
    resource.unassign()

  def __getitem__(self, idx: int) -> S:
    """Get a site by index."""
    return self.sites[idx]

  def __setitem__(self, idx: int, resource: Optional[Resource]):
    """Assign a resource to this carrier."""
    if resource is None:  # setting to None
      assigned_resource = self[idx].resource
      if assigned_resource is not None:
        self.unassign_child_resource(assigned_resource)
    else:
      self.assign_resource_to_site(resource, spot=idx)

  def __delitem__(self, idx: int):
    """Unassign a resource from this carrier."""
    assigned_resource = self[idx].resource
    if assigned_resource is not None:
      self.unassign_child_resource(assigned_resource)

  def get_resources(self) -> List[Resource]:
    """Get all resources, using self.__getitem__ (so that the location is within this carrier)."""
    all_resources = [site.resource for site in self.sites.values()]
    return [resource for resource in all_resources if resource is not None]

  def __eq__(self, other):
    return super().__eq__(other) and self.sites == other.sites

  def get_free_sites(self) -> List[S]:
    return [site for site in self.sites.values() if site.resource is None]


class TipCarrier(Carrier):
  r"""Base class for tip carriers.
  Name prefix: 'TIP\_'
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    sites: Optional[Dict[int, ResourceHolder]] = None,
    category="tip_carrier",
    model: Optional[str] = None,
  ):
    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      sites,
      category=category,
      model=model,
    )


class PlateHolder(ResourceHolder):
  """A single site within a plate carrier."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    pedestal_size_z: float = None,  # type: ignore
    child_location=Coordinate.zero(),
    category="plate_holder",
    model: Optional[str] = None,
  ):
    super().__init__(
      name, size_x, size_y, size_z, category=category, model=model, child_location=child_location
    )
    if pedestal_size_z is None:
      raise ValueError(
        "pedestal_size_z must be provided. See "
        "https://docs.pylabrobot.org/resources/resource-holder/plate-holder.html#pedestal-z-height for more "
        "information."
      )

    self.pedestal_size_z = pedestal_size_z
    self.resource: Optional[Plate]  # fix type
    # TODO: add self.pedestal_2D_offset if necessary in the future

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate] = None,
    reassign: bool = True,
  ):
    if isinstance(resource, ResourceStack):
      if not resource.direction == "z":
        raise ValueError("ResourceStack assigned to PlateHolder must have direction 'z'")
      if not all(isinstance(c, Plate) for c in resource.children):
        raise TypeError(
          "If a ResourceStack is assigned to a PlateHolder, the items "
          + f"must be Plates, not {type(resource.children[-1])}"
        )
    if isinstance(resource, Plate) and resource.plate_type != "skirted":
      raise ValueError("PlateHolder can only store plates that are skirted")
    return super().assign_child_resource(resource, location, reassign)

  def _get_sinking_depth(self, resource: Resource) -> Coordinate:
    def get_plate_sinking_depth(plate: Plate):
      # Sanity check for equal well clearances / dz
      well_dz_set = {
        round(well.location.z, 2)
        for well in plate.get_all_children()
        if well.category == "well" and well.location is not None
      }
      assert len(well_dz_set) == 1, "All wells must have the same z location"
      well_dz = well_dz_set.pop()
      # Plate "sinking" logic based on well dz to pedestal relationship
      pedestal_size_z = abs(self.pedestal_size_z)
      z_sinking_depth = min(pedestal_size_z, well_dz)
      return z_sinking_depth

    z_sinking_depth = 0.0
    if isinstance(resource, Plate):
      z_sinking_depth = get_plate_sinking_depth(resource)
    elif isinstance(resource, ResourceStack) and len(resource.children) > 0:
      first_child = resource.children[0]
      if isinstance(first_child, Plate):
        z_sinking_depth = get_plate_sinking_depth(first_child)

      # TODO #246 - _get_sinking_depth should not handle callbacks
      resource.register_did_assign_resource_callback(self._update_resource_stack_location)
      self.register_did_unassign_resource_callback(self._deregister_resource_stack_callback)
    return -Coordinate(z=z_sinking_depth)

  def get_default_child_location(self, resource: Resource) -> Coordinate:
    return super().get_default_child_location(resource) + self._get_sinking_depth(resource)

  def _update_resource_stack_location(self, resource: Resource):
    """Callback called when the lowest resource on a ResourceStack changes. Since the location of
    the lowest resource on the stack wrt the ResourceStack is always 0,0,0, we need to update the
    location of the ResourceStack itself to make sure we take into account sinking of the plate.

    Args:
      resource: The Resource on the ResourceStack that was assigned.
    """
    resource_stack = resource.parent
    assert isinstance(resource_stack, ResourceStack)
    if resource_stack.children[0] == resource:
      resource_stack.location = self.get_default_child_location(resource)

  def _deregister_resource_stack_callback(self, resource: Resource):
    """Callback called when a ResourceStack (or child) is unassigned from this PlateHolder."""
    if isinstance(resource, ResourceStack):  # the ResourceStack itself is unassigned
      resource.deregister_did_assign_resource_callback(self._update_resource_stack_location)

  def serialize(self) -> dict:
    return {
      **super().serialize(),
      "pedestal_size_z": self.pedestal_size_z,
    }


class PlateCarrier(Carrier):
  r"""Base class for plate carriers.
  Name prefix: 'PLT\_'
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    sites: Optional[Dict[int, PlateHolder]] = None,
    category="plate_carrier",
    model: Optional[str] = None,
  ):
    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      sites,
      category=category,
      model=model,
    )
    self.sites: Dict[int, PlateHolder] = sites or {}  # fix type

  def summary(self) -> str:
    """Return a summary of the carrier's sites and their contents."""

    def create_pretty_table(header, *columns) -> str:
      col_widths = [
        max(len(str(item)) for item in [header[i]] + list(columns[i])) for i in range(len(header))
      ]

      def format_row(row, border="|") -> str:
        return (
          f"{border} "
          + " | ".join(f"{str(row[i]).ljust(col_widths[i])}" for i in range(len(row)))
          + f" {border}"
        )

      def separator_line(cross: str = "+", line: str = "-") -> str:
        return cross + cross.join(line * (width + 2) for width in col_widths) + cross

      table = []
      table.append(separator_line())  # Top border
      table.append(format_row(header))
      table.append(separator_line())  # Header separator
      for row in zip(*columns):
        table.append(format_row(row))
      table.append(separator_line())  # Bottom border
      return "\n".join(table)

    indices = sorted(self.sites.keys())
    header = ["Site", "Content"]
    site_numbers = list(reversed([str(i) for i in indices]))
    site_resources = list(reversed([self.sites[i].resource for i in indices]))
    site_contents = [r.name if r is not None else "<empty>" for r in site_resources]
    return create_pretty_table(header, site_numbers, site_contents)


class MFXCarrier(Carrier[ResourceHolder]):
  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    sites: Optional[Dict[int, ResourceHolder]] = None,
    category="mfx_carrier",
    model: Optional[str] = None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      sites=sites,
      category=category,
      model=model,
    )


class TubeCarrier(Carrier):
  r"""Base class for tube/sample carriers.
  Name prefix: 'SMP\_'
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    sites: Optional[Dict[int, ResourceHolder]] = None,
    category="tube_carrier",
    model: Optional[str] = None,
  ):
    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      sites,
      category=category,
      model=model,
    )


class TroughCarrier(Carrier):
  r"""Base class for reagent/trough carriers.
  Name prefix: 'RGT\_'
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    sites: Optional[Dict[int, ResourceHolder]] = None,
    category="trough_carrier",
    model: Optional[str] = None,
  ):
    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      sites,
      category=category,
      model=model,
    )


T = TypeVar("T", bound=ResourceHolder)


def create_resources(
  klass: Type[T],
  locations: List[Coordinate],
  resource_size_x: List[Union[float, int]],
  resource_size_y: List[Union[float, int]],
  resource_size_z: Optional[List[Union[float, int]]] = None,
  name_prefix: Optional[str] = None,
  **kwargs,
) -> Dict[int, T]:
  """Create a list of resource with the given sizes and locations.

  Args:
    klass: The class of the resources.
    locations: The locations of the resources.
    resource_size_x: The x size of the resources.
    resource_size_y: The y size of the resources.
    resource_size_z: The z size of the resources. If None, it will be set to 0.
    name_prefix: names of the resources will be f"{name_prefix}-{idx}" if name_prefix is not None,
      else f"{klass.__name__}_{idx}".
  """
  # TODO: should be possible to merge with create_equally_spaced_y

  if resource_size_z is None:
    resource_size_z = [0] * len(locations)

  sites = {}
  for idx, (location, x, y, z) in enumerate(
    zip(locations, resource_size_x, resource_size_y, resource_size_z)
  ):
    site = klass(
      name=f"{name_prefix}-{idx}" if name_prefix else f"{klass.__name__}_{idx}",
      size_x=x,
      size_y=y,
      size_z=z,
      **kwargs,
    )
    site.location = location
    sites[idx] = site
  return sites


def create_homogeneous_resources(
  klass: Type[T],
  locations: List[Coordinate],
  resource_size_x: float,
  resource_size_y: float,
  resource_size_z: Optional[float] = None,
  name_prefix: Optional[str] = None,
  **kwargs,
) -> Dict[int, T]:
  """Create a list of resources with the same size at specified locations."""
  # TODO: should be possible to merge with create_equally_spaced_y

  n = len(locations)
  return create_resources(
    klass=klass,
    locations=locations,
    resource_size_x=[resource_size_x] * n,
    resource_size_y=[resource_size_y] * n,
    resource_size_z=[resource_size_z] * n if resource_size_z is not None else None,
    name_prefix=name_prefix,
    **kwargs,
  )



import os
import unittest

import pylabrobot
from pylabrobot.pumps.calibration import PumpCalibration

plr_directory = os.path.join(pylabrobot.__path__[0], "testing", "test_data")


class TestCalibration(unittest.TestCase):
  """Tests for the PumpCalibration class."""

  def test_load_calibration_value(self):
    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration(1.0)

  def test_load_from_txt(self):
    txt_path = os.path.join(plr_directory, "test_calibration.txt")
    with self.assertRaises(NotImplementedError):
      PumpCalibration.load_calibration(txt_path)

  def test_load_from_json(self):
    json_list_path = os.path.join(plr_directory, "test_calibration_list.json")
    calibration = PumpCalibration.load_calibration(json_list_path)
    self.assertEqual(calibration[0], 1)
    self.assertEqual(calibration[1], 1)

    json_dict_path = os.path.join(plr_directory, "test_calibration_dict.json")
    calibration = PumpCalibration.load_calibration(json_dict_path)
    self.assertEqual(calibration[0], 1.0)
    self.assertEqual(calibration[1], 1.0)

    json_null_path = os.path.join(plr_directory, "test_calibration_null.json")
    self.assertRaises(TypeError, PumpCalibration.load_calibration, json_null_path)

  def test_load_from_csv(self):
    csv_path = os.path.join(plr_directory, "test_calibration.csv")
    calibration = PumpCalibration.load_calibration(csv_path)
    self.assertEqual(calibration[0], 1.0)
    self.assertEqual(calibration[1], 1.0)

    csv_path = os.path.join(plr_directory, "test_calibration_three_columns.csv")
    self.assertRaises(ValueError, PumpCalibration.load_calibration, csv_path)

  def test_load_from_dict(self):
    calibration = PumpCalibration.load_calibration({0: 1.0, 1: 2.0})
    self.assertEqual(calibration[0], 1.0)
    self.assertEqual(calibration[1], 2.0)

    calibration = PumpCalibration.load_calibration({1: 1.0, 0: 2.0})
    self.assertEqual(calibration[0], 2.0)
    self.assertEqual(calibration[1], 1.0)

  def test_load_from_dict_errors(self):
    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration({0: 1.0, 1: 2.0, 3: 3.0})  # missing key 2

    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration({0: -1.0, 1: 2.0, 2: 3.0})  # negative value

    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration({-1: 1.0, 0: 1.0, 1: 2.0})  # negative key

    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration({2: 1.0, 5: 1.0, 1: 2.0})  # missing key 0

    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration({2: 1.0, 3: 1.0, 4: 2.0})  # missing key 0, 1

  def test_load_from_list(self):
    test_list = [1.0, 2.0]
    calibration = PumpCalibration.load_calibration(test_list)
    self.assertEqual(calibration[0], 1.0)
    self.assertEqual(calibration[1], 2.0)

  def test_load_from_value(self):
    test_value = 1.0
    calibration = PumpCalibration.load_calibration(test_value, 2)
    self.assertEqual(calibration[0], 1.0)
    self.assertEqual(calibration[1], 1.0)

  def test_calibration_mode(self):
    calibration = PumpCalibration([1.0, 2.0], calibration_mode="revolutions")
    self.assertEqual(calibration.calibration_mode, "revolutions")

    calibration = PumpCalibration.load_calibration({0: 1.0, 1: 2.0}, calibration_mode="revolutions")
    self.assertEqual(calibration.calibration_mode, "revolutions")

    calibration = PumpCalibration.load_calibration(1.0, 2)
    self.assertEqual(calibration.calibration_mode, "duration")

  def test_calibration_mode_errors(self):
    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration(
        [1.0, 2.0],
        calibration_mode="invalid",  # type: ignore[arg-type]
      )

    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration(
        {0: 1.0, 1: 2.0},
        calibration_mode="invalid",  # type: ignore[arg-type]
      )

    with self.assertRaises(ValueError):
      PumpCalibration.load_calibration(1.0, 2, calibration_mode="invalid")  # type: ignore[arg-type]



"""Corning plates."""

from pylabrobot.resources.height_volume_functions import (
  calculate_liquid_height_in_container_2segments_square_vbottom,
  calculate_liquid_volume_container_2segments_square_vbottom,
)
from pylabrobot.resources.plate import Lid, Plate
from pylabrobot.resources.utils import create_ordered_items_2d
from pylabrobot.resources.well import (
  CrossSectionType,
  Well,
  WellBottomType,
)

# # # # # # # # # # Cor_96_wellplate_360ul_Fb # # # # # # # # # #

# Well tapers from 6.35 mm (bottom) to 6.86 mm (top) over 10.67 mm depth.
_cor_96_wellplate_360ul_Fb_height_volume_data = {
  0.0: 0.0,
  0.45: 20.0,  # "dead volume" to cover the full flat bottom and ensure LLD detectability
  1.69: 50.0,
  3.22: 100.0,
  4.72: 150.0,
  6.19: 200.0,
  7.72: 250.0,
  8.99: 300.0,
  10.62: 350.0,
  10.95: 360.0,
}


def Cor_96_wellplate_360ul_Fb(name: str, with_lid: bool = False) -> Plate:
  """
  Corning cat. no.s: 3603
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2b/UK/en/Microplates/
    Assay-Microplates/96-Well-Microplates/Corning%C2%AE-96-well-Black-Clear-and-White-Clear-
    Bottom-Polystyrene-Microplates/p/3603
  - brand: Corning
  - distributor: (Fisher Scientific, 10530753)
  - material: Polypropylene
  - tc_treated: Yes
  - tech_drawing: tech_drawings/Cor_96_wellplate_360ul_Fb.pdf
  - sterile: yes
  - notes:
      - Well bottom (i.e. material_z_thickness) is 60% thinner than conventional polystyrene
        microplates, resulting in lower background fluorescence & enabling readings down to 340 nm.
      - Opaque walls prevent well-to-well cross-talk.
  """

  # This used to be Cos_96_EZWash in the Esvelt lab

  return Plate(
    name=name,
    size_x=127.76,
    size_y=85.48,
    size_z=14.2,
    lid=Cor_96_wellplate_360ul_Fb_Lid(name=name + "_lid") if with_lid else None,
    model="Cor_96_wellplate_360ul_Fb",
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=10.87,  # 14.3-6.86/2
      dy=7.77,  # 11.2-6.86/2
      dz=3.03,
      item_dx=9.0,
      item_dy=9.0,
      size_x=6.86,  # top
      size_y=6.86,  # top
      size_z=10.67,
      material_z_thickness=0.5,
      bottom_type=WellBottomType.FLAT,
      cross_section_type=CrossSectionType.CIRCLE,
      max_volume=360,
      height_volume_data=_cor_96_wellplate_360ul_Fb_height_volume_data,
    ),
  )


def Cor_96_wellplate_360ul_Fb_Lid(name: str) -> Lid:
  """
  - brand: Corning
  """

  return Lid(
    name=name,
    size_x=127.76,
    size_y=85.48,
    size_z=8.9,  # measure the total z height
    nesting_z_height=7.6,  # measure overlap between lid and plate
    model="Cor_96_wellplate_360ul_Fb_Lid",
  )


# Previous names in PLR:
def Cos_96_EZWash(name: str, with_lid: bool = False) -> Plate:
  raise ValueError("Deprecated. You probably want to use Cor_96_wellplate_360ul_Fb instead.")


# # # # # # # # # # Cor_96_wellplate_2mL_Vb # # # # # # # # # #


def Cor_96_wellplate_2mL_Vb(name: str, with_lid: bool = False) -> Plate:
  """
  Corning cat. no.: 3960
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2b/UK/en/Genomics-%26-
    Molecular-Biology/Automation-Consumables/Deep-Well-Plate/Corning%C2%AE-96-well-
    Polypropylene-Storage-Blocks/p/3960
  - brand: Corning
  - distributor: (Fisher Scientific, 10708212)
  - material: Polypropylene
  - sterile: yes
  - notes:
      - features uniform skirt heights for greater robotic gripping surface.
  """
  return Plate(
    name=name,
    size_x=127.0,
    size_y=86.0,
    size_z=43.5,
    lid=Cor_96_wellplate_2mL_Vb_Lid(name=name + "_lid") if with_lid else None,
    model="Cor_Cos_96_wellplate_2mL_Vb",
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=9.6,
      dy=7.0,
      dz=1.2,
      item_dx=9.0,
      item_dy=9.0,
      size_x=8.0,
      size_y=8.0,
      size_z=42.0,
      bottom_type=WellBottomType.V,
      material_z_thickness=1.25,
      cross_section_type=CrossSectionType.RECTANGLE,
      compute_volume_from_height=_compute_volume_from_height_Cor_96_wellplate_2mL_Vb,
      compute_height_from_volume=_compute_height_from_volume_Cor_96_wellplate_2mL_Vb,
    ),
  )


def Cor_96_wellplate_2mL_Vb_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")


# Volume-height functions
def _compute_volume_from_height_Cor_96_wellplate_2mL_Vb(
  h: float,
) -> float:
  if h > 44.1:  # 5% tolerance
    raise ValueError(f"Height {h} is too large for Cor_96_wellplate_2mL_Vb")
  return calculate_liquid_volume_container_2segments_square_vbottom(
    x=7.8, y=7.8, h_pyramid=4.0, h_cube=38.0, liquid_height=h
  )


def _compute_height_from_volume_Cor_96_wellplate_2mL_Vb(
  liquid_volume: float,
):
  if liquid_volume > 2_100:  # 5% tolerance
    raise ValueError(f"Volume {liquid_volume} is too large for Cor_96_wellplate_2mL_Vb")
  return round(
    calculate_liquid_height_in_container_2segments_square_vbottom(
      x=7.8,
      y=7.8,
      h_pyramid=4.0,
      h_cube=38.0,
      liquid_volume=liquid_volume,
    ),
    3,
  )


# Previous names in PLR:
def Cos_96_DWP_2mL_Vb(name: str, with_lid: bool = False) -> Plate:
  raise NotImplementedError(
    "This function is deprecated and will be removed in a future version."
    " Use 'Cor_96_wellplate_2mL_Vb' instead."
  )


def Cos_96_wellplate_2mL_Vb(name: str, with_lid: bool = False) -> Plate:
  raise NotImplementedError("deprecated. use Cor_96_wellplate_2mL_Vb instead")



import json
import math
import os
import urllib.request
from typing import Dict, List, cast

from pylabrobot.resources import Coordinate, Tip, TipRack, TipSpot
from pylabrobot.resources.carrier import PlateHolder
from pylabrobot.resources.resource_holder import ResourceHolder
from pylabrobot.resources.tube_rack import TubeRack


def _download_file(url: str, local_path: str) -> bytes:
  with urllib.request.urlopen(url) as response, open(local_path, "wb") as out_file:
    data = response.read()
    out_file.write(data)
    return data  # type: ignore


def _download_ot_resource_file(ot_name: str, force_download: bool):
  """Download an Opentrons tip rack definition file from GitHub.

  Args:
    ot_name: The name of the tip rack, like "opentrons_96_tiprack_300ul".

  Returns:
    The labware definition as a dictionary.
  """
  url = f"https://raw.githubusercontent.com/Opentrons/opentrons/5b51a98ce736b2bb5aff780bf3fdf91941a038fa/shared-data/labware/definitions/2/{ot_name}/1.json"
  if os.path.exists("/tmp"):
    path = f"/tmp/{ot_name}.json"  # only works with linux/mac systems
  else:
    path = f"C:/Windows/Temp/{ot_name}.json"
  if force_download or not os.path.exists(path):
    data = _download_file(url=url, local_path=path)
  else:
    with open(path, "rb") as f:
      data = f.read()
  return json.loads(data)


def load_ot_tip_rack(
  ot_name: str, plr_resource_name: str, with_tips: bool = True, force_download: bool = False
) -> TipRack:
  """Convert an Opentrons tip rack definition file to a PyLabRobot TipRack resource."""

  data = _download_ot_resource_file(ot_name=ot_name, force_download=force_download)

  display_category = data["metadata"]["displayCategory"]
  if not display_category == "tipRack":
    raise ValueError("Not a tip rack definition file.")

  items = data["ordering"]
  wells: List[TipSpot] = []

  for column in items:
    for item in column:
      well_data = data["wells"][item]

      assert well_data["shape"] == "circular", "We assume all tip racks are circular."
      diameter = well_data["diameter"]
      well_size_x = well_size_y = round(diameter / math.sqrt(2), 3)

      # closure
      def make_tip(name: str) -> Tip:
        return Tip(
          name=name,
          total_tip_length=data["parameters"]["tipLength"],
          has_filter="Filter" in data["metadata"]["displayName"],
          maximal_volume=well_data["totalLiquidVolume"],
          fitting_depth=data["parameters"]["tipOverlap"],
        )

      tip_spot = TipSpot(
        name=item,
        size_x=well_size_x,
        size_y=well_size_y,
        make_tip=make_tip,
      )
      tip_spot.location = Coordinate(
        x=well_data["x"] - well_size_x / 2,
        y=well_data["y"] - well_size_y / 2,
        z=well_data["z"],
      )
      wells.append(tip_spot)

  ordering = data["ordering"]
  flattened_ordering = [item for sublist in ordering for item in sublist]
  ordered_items = dict(zip(flattened_ordering, wells))

  tr = TipRack(
    name=plr_resource_name,
    size_x=data["dimensions"]["xDimension"],
    size_y=data["dimensions"]["yDimension"],
    size_z=data["dimensions"]["zDimension"],
    ordered_items=cast(Dict[str, TipSpot], ordered_items),
    model=data["metadata"]["displayName"],
  )
  if with_tips:
    tr.fill()
  else:
    tr.empty()
  return tr


def load_ot_tube_rack(
  ot_name: str, plr_resource_name: str, force_download: bool = False
) -> TubeRack:
  """Convert an Opentrons tube rack definition file to a PyLabRobot TubeRack resource."""

  data = _download_ot_resource_file(ot_name=ot_name, force_download=force_download)

  display_category = data["metadata"]["displayCategory"]
  if display_category not in {"tubeRack", "aluminumBlock"}:
    raise ValueError("Not a tube rack definition file.")

  items = data["ordering"]
  wells: List[ResourceHolder] = []

  for column in items:
    for item in column:
      well_data = data["wells"][item]

      assert well_data["shape"] == "circular", "We assume all tip racks are circular."
      diameter = well_data["diameter"]
      well_size_x = well_size_y = round(diameter / math.sqrt(2), 3)
      well_size_z = well_data["depth"]

      resource_holder = ResourceHolder(
        name=item,
        size_x=well_size_x,
        size_y=well_size_y,
        size_z=well_size_z,
      )
      resource_holder.location = Coordinate(
        x=well_data["x"] - well_size_x / 2,
        y=well_data["y"] - well_size_y / 2,
        z=well_data["z"],
      )
      wells.append(resource_holder)

  ordering = data["ordering"]
  flattened_ordering = [item for sublist in ordering for item in sublist]
  ordered_items = dict(zip(flattened_ordering, wells))

  return TubeRack(
    name=plr_resource_name,
    size_x=data["dimensions"]["xDimension"],
    size_y=data["dimensions"]["yDimension"],
    size_z=data["dimensions"]["zDimension"],
    ordered_items=cast(Dict[str, ResourceHolder], ordered_items),
    model=data["metadata"]["displayName"],
  )


def load_ot_plate_holder(
  ot_name: str, plr_resource_name: str, z_offset: float, force_download: bool = False
) -> PlateHolder:
  """Convert an Opentrons adapter definition file to a PyLabRobot PlateHolder resource."""

  data = _download_ot_resource_file(ot_name=ot_name, force_download=force_download)

  display_category = data["metadata"]["displayCategory"]
  if display_category not in {"adapter", "aluminumBlock"}:
    raise ValueError("Not a plate adapter definition file.")

  location = data["cornerOffsetFromSlot"]

  return PlateHolder(
    name=plr_resource_name,
    size_x=data["dimensions"]["xDimension"],
    size_y=data["dimensions"]["yDimension"],
    size_z=data["dimensions"]["zDimension"],
    child_location=Coordinate(location["x"], location["y"], z_offset),
    pedestal_size_z=0,
    model=data["metadata"]["displayName"],
  )



import asyncio
import ctypes
import json
import logging
import math
import os
import time
import warnings
from typing import Optional

from pylabrobot.io.ftdi import FTDI

from .backend import CentrifugeBackend, LoaderBackend
from .standard import LoaderNoPlateError

logger = logging.getLogger(__name__)


class Access2Backend(LoaderBackend):
  def __init__(
    self,
    device_id: str,
    timeout: int = 60,
  ):
    """
    Args:
      device_id: The libftdi id for the loader. Find using
        `python3 -m pylibftdi.examples.list_devices`
    """
    self.io = FTDI(human_readable_device_name="Agilent Access2 Loader", device_id=device_id)
    self.timeout = timeout

  async def _read(self) -> bytes:
    x = b""
    r = None
    start = time.time()
    while r != b"" or x == b"":
      r = await self.io.read(1)
      x += r
      if r == b"":
        await asyncio.sleep(0.1)
      if x == b"" and (time.time() - start) > self.timeout:
        raise TimeoutError("No data received within the specified timeout period")
    return x

  async def send_command(self, command: bytes) -> bytes:
    logger.debug("[loader] Sending %s", command.hex())
    await self.io.write(command)
    return await self._read()

  async def setup(self):
    logger.debug("[loader] setup")

    await self.io.setup()
    await self.io.set_baudrate(115384)

    status = await self.get_status()
    if not status.startswith(bytes.fromhex("1105")):
      raise RuntimeError("Failed to get status")

    await self.send_command(bytes.fromhex("110500030014000072b1"))
    await self.send_command(bytes.fromhex("1105000300100000ae71"))
    await self.send_command(bytes.fromhex("110500070024040000008000be89"))
    await self.send_command(bytes.fromhex("11050007002404008000800063b1"))
    await self.send_command(bytes.fromhex("11050007002404000001800089b9"))
    await self.send_command(bytes.fromhex("1105000700240400800180005481"))
    await self.send_command(bytes.fromhex("110500070024040000024000c6bd"))
    await self.send_command(bytes.fromhex("1105000300400000f0bf"))
    await self.send_command(bytes.fromhex("1105000a004607000100000000020235bf"))
    # await self.send_command(bytes.fromhex("11050003002000006bd4"))
    await self.send_command(bytes.fromhex("1105000e00440b00000000000000007041020203c7"))
    # await self.send_command(bytes.fromhex("11050003002000006bd4"))

  async def stop(self):
    logger.debug("[loader] stop")
    await self.io.stop()

  def serialize(self):
    return {"io": self.io.serialize(), "timeout": self.timeout}

  async def get_status(self) -> bytes:
    logger.debug("[loader] get_status")
    return await self.send_command(bytes.fromhex("11050003002000006bd4"))

  async def park(self):
    logger.debug("[loader] park")
    await self.send_command(bytes.fromhex("1105000e00440b0000000000410000704103007539"))

  async def close(self):
    logger.debug("[loader] close")
    await self.send_command(bytes.fromhex("1105000a00420700010000803f02008c64"))

  async def open(self):
    logger.debug("[loader] open")
    await self.send_command(bytes.fromhex("1105000a0042070001000080bf0200b73e"))

  async def load(self):
    """only tested for 1cm plate, 3mm pickup height"""
    logger.debug("[loader] load")

    await self.send_command(bytes.fromhex("1105000a004607000100000000020235bf"))
    await self.send_command(bytes.fromhex("1105000e00440b000100004040000020410200a5cb"))

    # laser check
    r = await self.send_command(bytes.fromhex("1105000300500000b3dc"))
    if r == bytes.fromhex("1105000800510500000300000079f1"):
      raise LoaderNoPlateError("no plate found on stage")

    await self.send_command(bytes.fromhex("1105000a00460700018fc2b540020023dc"))
    await self.send_command(bytes.fromhex("1105000e00440b000200004040000020410300ee00"))
    await self.send_command(bytes.fromhex("1105000a004607000100000000020015fd"))
    await self.send_command(bytes.fromhex("1105000e00440b0000000040400000204102007d82"))

  async def unload(self):
    """only tested for 1cm plate, 3mm pickup height"""
    logger.debug("[loader] unload")

    await self.send_command(bytes.fromhex("1105000a004607000100000000020235bf"))
    await self.send_command(bytes.fromhex("1105000e00440b000200004040000020410200dd31"))

    # laser check
    r = await self.send_command(bytes.fromhex("1105000300500000b3dc"))
    if r == bytes.fromhex("1105000800510500000300000079f1"):
      raise LoaderNoPlateError("no plate found in centrifuge")

    await self.send_command(bytes.fromhex("1105000a00460700017b14b6400200d57a"))
    await self.send_command(bytes.fromhex("1105000e00440b00010000404000002041030096fa"))
    await self.send_command(bytes.fromhex("1105000a004607000100000000020015fd"))
    await self.send_command(bytes.fromhex("1105000e00440b00000000000000002041020056be"))
    # await self.send_command(bytes.fromhex("11050003002000006bd4"))


_vspin_bucket_calibrations_path = os.path.join(
  os.path.expanduser("~"),
  ".pylabrobot",
  "vspin_bucket_calibrations.json",
)


def _load_vspin_calibrations(device_id: str) -> Optional[int]:
  if not os.path.exists(_vspin_bucket_calibrations_path):
    warnings.warn(
      f"No calibration found for VSpin with device id {device_id}. "
      "Please set the bucket 1 position using `set_bucket_1_position_to_current` method after setup.",
      UserWarning,
    )
    return None
  with open(_vspin_bucket_calibrations_path, "r") as f:
    return json.load(f).get(device_id)  # type: ignore


def _save_vspin_calibrations(device_id, remainder: int):
  if os.path.exists(_vspin_bucket_calibrations_path):
    with open(_vspin_bucket_calibrations_path, "r") as f:
      data = json.load(f)
  else:
    data = {}
  data[device_id] = remainder
  os.makedirs(os.path.dirname(_vspin_bucket_calibrations_path), exist_ok=True)
  with open(_vspin_bucket_calibrations_path, "w") as f:
    json.dump(data, f)


FULL_ROTATION: int = 8000


bucket_1_not_set_error = RuntimeError(
  "Bucket 1 position not set. "
  "Please rotate the bucket to bucket 1 using VSpinBackend.go_to_position and "
  "then calling VSpinBackend.set_bucket_1_position_to_current."
)


class VSpinBackend(CentrifugeBackend):
  """Backend for the Agilent Centrifuge.
  Note that this is not a complete implementation."""

  def __init__(self, device_id: Optional[str] = None):
    """
    Args:
      device_id: The libftdi id for the centrifuge. Find using `python -m pylibftdi.examples.list_devices`
    """
    self.io = FTDI(human_readable_device_name="Agilent VSpin Centrifuge", device_id=device_id)
    self._bucket_1_remainder: Optional[int] = None
    # only attempt loading calibration if device_id is not None
    # if it is None, we will load it after setup when we can query the device id from the io
    if device_id is not None:
      self._bucket_1_remainder = _load_vspin_calibrations(device_id)

  async def setup(self):
    await self.io.setup()
    # TODO: add functionality where if robot has been initialized before nothing needs to happen
    for _ in range(3):
      await self.configure_and_initialize()
      await self._send_command(bytes.fromhex("aa002101ff21"))
    await self._send_command(bytes.fromhex("aa002101ff21"))
    await self._send_command(bytes.fromhex("aa01132034"))
    await self._send_command(bytes.fromhex("aa002102ff22"))
    await self._send_command(bytes.fromhex("aa02132035"))
    await self._send_command(bytes.fromhex("aa002103ff23"))
    await self._send_command(bytes.fromhex("aaff1a142d"))

    await self.io.set_baudrate(57600)
    await self.io.set_rts(True)
    await self.io.set_dtr(True)

    await self._send_command(bytes.fromhex("aa01121f32"))
    for _ in range(8):
      await self._send_command(bytes.fromhex("aa0220ff0f30"))
    await self._send_command(bytes.fromhex("aa0220df0f10"))
    await self._send_command(bytes.fromhex("aa0220df0e0f"))
    await self._send_command(bytes.fromhex("aa0220df0c0d"))
    await self._send_command(bytes.fromhex("aa0220df0809"))
    for _ in range(4):
      await self._send_command(bytes.fromhex("aa0226000028"))
    await self._send_command(bytes.fromhex("aa02120317"))
    for _ in range(5):
      await self._send_command(bytes.fromhex("aa0226200048"))
      await self._send_command(bytes.fromhex("aa0226000028"))
    await self.lock_door()

    await self._send_command(bytes.fromhex("aa0226000028"))

    await self._send_command(bytes.fromhex("aa0117021a"))
    await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
    await self._send_command(bytes.fromhex("aa0117041c"))
    await self._send_command(bytes.fromhex("aa01170119"))

    await self._send_command(bytes.fromhex("aa010b0c"))
    await self._send_command(bytes.fromhex("aa010001"))
    await self._send_command(bytes.fromhex("aa01e605006400000000003200e80301006e"))
    await self._send_command(bytes.fromhex("aa0194b61283000012010000f3"))
    await self._send_command(bytes.fromhex("aa01192842"))

    resp = 0x89
    while resp == 0x89:
      resp = (await self._get_positions_and_tachometer()).status

    # --- almost the same as go to position ---
    await self._send_command(bytes.fromhex("aa0117021a"))
    await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
    await self._send_command(bytes.fromhex("aa0117041c"))
    await self._send_command(bytes.fromhex("aa01170119"))

    await self._send_command(bytes.fromhex("aa010b0c"))
    await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
    new_position = (0).to_bytes(4, byteorder="little")  # arbitrary
    # rpm = 600,
    # acceleration = 75.09289617486338
    await self._send_command(
      bytes.fromhex("aa01d497") + new_position + bytes.fromhex("c3f52800d71a000049")
    )
    # -----------------------------------------

    resp = 0x08
    while resp != 0x09:
      resp = (await self._get_positions_and_tachometer()).status

    await self._send_command(bytes.fromhex("aa0117021a"))

    await self.lock_door()

    # If we have not set the calibration yet, load it now.
    if self._bucket_1_remainder is None:
      device_id = await self.io.get_serial()
      self._bucket_1_remainder = _load_vspin_calibrations(device_id)

  @property
  def bucket_1_remainder(self) -> int:
    if self._bucket_1_remainder is None:
      raise bucket_1_not_set_error
    return self._bucket_1_remainder

  async def set_bucket_1_position_to_current(self) -> None:
    """Set the current position as bucket 1 position and save calibration."""
    current_position = await self.get_position()
    device_id = await self.io.get_serial()
    remainder = await self.get_home_position() - current_position
    self._bucket_1_remainder = current_position % FULL_ROTATION
    _save_vspin_calibrations(device_id, remainder)

  async def get_bucket_1_position(self) -> int:
    """Get the bucket 1 position based on calibration.
    Normally it is the home position minus the remainder (calibration).
    The bucket 1 position must be greater than the current position, so we find
    the first position greater than the current position by adding full rotations if needed.
    """
    if self._bucket_1_remainder is None:
      raise bucket_1_not_set_error
    home_position = await self.get_home_position()
    bucket_1_position_mod_full_rotation = home_position - self.bucket_1_remainder
    # first number after current position that matches bucket 1 position mod FULL_ROTATION
    current_position = await self.get_position()
    bucket_1_position = (
      FULL_ROTATION
      * math.floor((current_position - bucket_1_position_mod_full_rotation) / FULL_ROTATION + 1)
      + bucket_1_position_mod_full_rotation
    )
    return bucket_1_position

  async def stop(self):
    await self.configure_and_initialize()
    await self.io.stop()

  class _StatusPositionTachometer(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
      ("status", ctypes.c_uint8),
      ("current_position", ctypes.c_uint32),
      ("unknown1", ctypes.c_uint8),
      ("tachometer", ctypes.c_int16),
      ("unknown2", ctypes.c_uint8),
      ("home_position", ctypes.c_uint32),
      ("checksum", ctypes.c_uint8),
    ]

  async def _get_positions_and_tachometer(self) -> _StatusPositionTachometer:
    """Returns 14 bytes

    Example:
      11 22 25 00 00 4f 00 00 18 e0 05 00 00 a4
                                             ^^ checksum
                                 ^^ ^^ ^^ ^^ home position
                              ^^ ? (probably binary status objects)
                        ^^ ^^ tachometer
                     ^^ ? (probably binary status objects)
         ^^ ^^ ^^ ^^ current position
      ^^
      - First byte (index 0):
        - 11 = 0b0001011 = idle
        - 13 = 0b0001101 = unknown
        - 08 = 0b0001000 = spinning
        - 09 = 0b0001001 = also spinning but different
        - 19 = 0b0010011 = unknown
        - 88 = 0b1011000 = unknown
        - 89 = 0b1011001 = unknown
      - 10th to 13th byte (index 9-12) = Homing Position
      - Last byte (index 13) = checksum
    """
    resp = await self._send_command(bytes.fromhex("aa010e0f"))
    if len(resp) == 0:
      raise IOError("Empty status from centrifuge")
    return VSpinBackend._StatusPositionTachometer.from_buffer_copy(resp)

  async def get_position(self) -> int:
    return (await self._get_positions_and_tachometer()).current_position  # type: ignore

  async def get_tachometer(self) -> int:
    """current speed in rpm"""
    tack_to_rpm = -14.69320388  # R^2 = 0.9999 when spinning, but not specific at single-digit RPM
    return (await self._get_positions_and_tachometer()).tachometer * tack_to_rpm  # type: ignore

  async def get_home_position(self) -> int:
    """changes during a run, but the bucket 1 position relative to it does not"""
    return (await self._get_positions_and_tachometer()).home_position  # type: ignore

  async def _get_status(self):
    """
    examples:
    - 0080d0015
    - 0080f0015
    """

    resp = await self._send_command(bytes.fromhex("aa020e10"))
    if len(resp) == 0:
      raise IOError("Empty status from centrifuge. Is the machine on?")
    return resp

  async def get_bucket_locked(self) -> bool:
    resp = await self._get_status()
    return resp[2] & 0b0001 != 0  # type: ignore

  async def get_door_open(self) -> bool:
    resp = await self._get_status()
    return resp[2] & 0b0010 != 0  # type: ignore

  async def get_door_locked(self) -> bool:
    resp = await self._get_status()
    return resp[2] & 0b0100 == 0  # type: ignore

  # Centrifuge communication: read_resp, send

  async def _read_resp(self, timeout: float = 20) -> bytes:
    """Read a response from the centrifuge. If the timeout is reached, return the data that has
    been read so far."""
    data = b""
    end_byte_found = False
    start_time = time.time()

    while True:
      chunk = await self.io.read(25)
      if chunk:
        data += chunk
        end_byte_found = data[-1] == 0x0D
        if len(chunk) < 25 and end_byte_found:
          break
      else:
        if end_byte_found or time.time() - start_time > timeout:
          break
        await asyncio.sleep(0.0001)

    logger.debug("Read %s", data.hex())
    return data

  async def _send_command(self, cmd: bytes, read_timeout=0.2) -> bytes:
    written = await self.io.write(bytes(cmd))

    if written != len(cmd):
      raise RuntimeError("Failed to write all bytes")
    return await self._read_resp(timeout=read_timeout)

  async def configure_and_initialize(self):
    await self.set_configuration_data()
    await self.initialize()

  async def set_configuration_data(self):
    """Set the device configuration data."""
    await self.io.set_latency_timer(16)
    await self.io.set_line_property(bits=8, stopbits=1, parity=0)
    await self.io.set_flowctrl(0)
    await self.io.set_baudrate(19200)

  async def initialize(self):
    await self.io.write(b"\x00" * 20)
    for i in range(33):
      packet = b"\xaa" + bytes([i & 0xFF, 0x0E, 0x0E + (i & 0xFF)]) + b"\x00" * 8
      await self.io.write(packet)
    await self._send_command(bytes.fromhex("aaff0f0e"))

  # Centrifuge operations

  async def open_door(self):
    if await self.get_door_open():
      return
    # used to be:                           aa022600072f
    await self._send_command(bytes.fromhex("aa022600062e"))  # same as unlock door

    # we can't tell when the door is fully open, so we just wait a bit
    await asyncio.sleep(4)

  async def close_door(self):
    if not (await self.get_door_open()):
      return
    # used to be:                           aa022600052d
    await self._send_command(bytes.fromhex("aa022600042c"))  # same as unlock door
    # we can't tell when the door is fully closed, so we just wait a bit
    await asyncio.sleep(2)

  async def lock_door(self):
    if await self.get_door_open():
      raise RuntimeError("Cannot lock door while it is open.")
    if await self.get_door_locked():
      return
    # used to be                            aa0226000129
    await self._send_command(bytes.fromhex("aa0226000028"))

  async def unlock_door(self):
    if not await self.get_door_locked():
      return
    # used to be                            aa022600052d
    await self._send_command(bytes.fromhex("aa022600042c"))  # same as close door

  async def lock_bucket(self):
    if await self.get_bucket_locked():
      return
    await self._send_command(bytes.fromhex("aa022600072f"))

  async def unlock_bucket(self):
    if not await self.get_bucket_locked():
      return
    await self._send_command(bytes.fromhex("aa022600062e"))  # same as open door

  async def go_to_bucket1(self):
    await self.go_to_position(await self.get_bucket_1_position())

  async def go_to_bucket2(self):
    await self.go_to_position(await self.get_bucket_1_position() + FULL_ROTATION // 2)

  async def go_to_position(self, position: int):
    await self.close_door()
    await self.lock_door()

    position_bytes = position.to_bytes(4, byteorder="little")
    byte_string = bytes.fromhex("aa01d497") + position_bytes + bytes.fromhex("c3f52800d71a0000")
    sum_byte = (sum(byte_string) - 0xAA) & 0xFF
    byte_string += sum_byte.to_bytes(1, byteorder="little")
    await self._send_command(bytes.fromhex("aa0226000028"))
    await self._send_command(bytes.fromhex("aa0117021a"))
    await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
    await self._send_command(bytes.fromhex("aa0117041c"))
    await self._send_command(bytes.fromhex("aa01170119"))
    await self._send_command(bytes.fromhex("aa010b0c"))
    await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
    await self._send_command(byte_string)

    # await self._send_command(bytes.fromhex("aa0117021a"))
    while (
      abs(await self.get_position() - position) > 10
    ):  # 10 tacks tolerance (10/8000 * 360 = 0.45 degrees)
      await asyncio.sleep(0.1)
    await self.open_door()

  @staticmethod
  def g_to_rpm(g: float) -> int:
    # https://en.wikipedia.org/wiki/Centrifugation#Mathematical_formula
    r = 10
    rpm = int((g / (1.118 * 10**-5 * r)) ** 0.5)
    return rpm

  async def spin(
    self,
    g: float = 500,
    duration: float = 60,
    acceleration: float = 0.8,
    deceleration: float = 0.8,
  ) -> None:
    """Start a spin cycle. spin spin spin spin

    Args:
      g: relative centrifugal force, also known as g-force
      duration: time in seconds spent at speed (g)
      acceleration: 0-1 of total acceleration
      deceleration: 0-1 of total deceleration
    """

    if acceleration <= 0 or acceleration > 1:
      raise ValueError("Acceleration must be within 0-1.")
    if deceleration <= 0 or deceleration > 1:
      raise ValueError("Deceleration must be within 0-1.")
    if g < 1 or g > 1000:
      raise ValueError("G-force must be within 1-1000")
    if duration < 1:
      raise ValueError("Spin time must be at least 1 second")

    if await self.get_door_open():
      await self.close_door()
    if not await self.get_door_locked():
      await self.lock_door()
    if await self.get_bucket_locked():
      await self.unlock_bucket()

    # 1 - compute the final position
    rpm = VSpinBackend.g_to_rpm(g)

    # compute the distance traveled during the acceleration period
    # distance = 1/2 * v^2 / a. area under 0 to t (triangle). t = a/v_max
    # 12903.2 ticks/s^2 is 100% acceleration
    acceleration_ticks_per_second2 = 12903.2 * acceleration
    rounds_per_second = rpm / 60
    ticks_per_second = rounds_per_second * 8000
    distance_during_acceleration = int(0.5 * (ticks_per_second**2) / acceleration_ticks_per_second2)

    # compute the distance traveled at speed
    distance_at_speed = ticks_per_second * duration

    current_position = await self.get_position()
    final_position = int(current_position + distance_during_acceleration + distance_at_speed)

    if final_position > 2**32 - 1:
      # this is almost 3 hours of spinning at 3000 rpm (max speed),
      # so we assume nobody will ever hit this.
      raise NotImplementedError(
        "We don't know what happens if the destination position exceeds 2^32-1. "
        "Please report this issue on discuss.pylabrobot.org."
      )

    # 2 - send "go to position" command with computed final position and rpm
    position_b = final_position.to_bytes(4, byteorder="little")
    rpm_b = int(rpm * 4473.925).to_bytes(4, byteorder="little")
    acceleration_b = int(9.15 * 100 * acceleration).to_bytes(4, byteorder="little")

    byte_string = bytes.fromhex("aa01d497") + position_b + rpm_b + acceleration_b
    checksum = (sum(byte_string) - 0xAA) & 0xFF
    byte_string += checksum.to_bytes(1, byteorder="little")

    await self._send_command(bytes.fromhex("aa0226000028"))
    await self._send_command(bytes.fromhex("aa0117021a"))
    await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
    await self._send_command(bytes.fromhex("aa0117041c"))
    await self._send_command(bytes.fromhex("aa01170119"))
    await self._send_command(bytes.fromhex("aa010b0c"))
    await self._send_command(bytes.fromhex("aa01e60500640000000000fd00803e01000c"))

    await self._send_command(byte_string)

    # 3 - wait for acceleration to the set rpm
    # we also check the position to avoid waiting forever if the speed is not reached (e.g. short spin...)
    while await self.get_tachometer() < rpm * 0.95 and await self.get_position() < final_position:
      await asyncio.sleep(0.1)

    # 4 - once the speed is reached, compute the position at which to start deceleration
    # this is different than computed above, because above we assumed constant acceleration from 0 to rpm.
    # however, in reality there is jerk and the acceleration is not constant, so we have to adjust as we go.
    # this is what the vendor software does too.
    # if we are already past that position, we skip this part.
    if await self.get_position() < final_position:
      decel_start_position = await self.get_position() + distance_at_speed

      # then wait until we reach that position
      while await self.get_position() < decel_start_position:
        await asyncio.sleep(0.1)

    # 5 - send deceleration command
    await self._send_command(bytes.fromhex("aa01e60500640000000000fd00803e01000c"))
    # aa0194b600000000dc02000029: decel at 80
    # aa0194b6000000000a03000058: decel at 85
    # aa0194b61283000012010000f3: used in setup (30%)
    decc = int(9.15 * 100 * deceleration).to_bytes(2, byteorder="little")
    decel_command = bytes.fromhex("aa0194b600000000") + decc + bytes.fromhex("0000")
    decel_command += ((sum(decel_command) - 0xAA) & 0xFF).to_bytes(1, byteorder="little")
    await self._send_command(decel_command)

    await asyncio.sleep(2)

    # 6 - reset position back to 0ish
    # this part is aneeded because otherwise calling go_to_position will not work after
    async def _reset_to_zero():
      await self._send_command(bytes.fromhex("aa0117021a"))
      await self._send_command(bytes.fromhex("aa01e6c800b00496000f004b00a00f050007"))
      await self._send_command(bytes.fromhex("aa0117041c"))
      await self._send_command(bytes.fromhex("aa01170119"))
      await self._send_command(bytes.fromhex("aa010b0c"))
      await self._send_command(bytes.fromhex("aa010001"))  # set position back to 0 (exactly)
      await self._send_command(bytes.fromhex("aa01e605006400000000003200e80301006e"))
      await self._send_command(bytes.fromhex("aa0194b61283000012010000f3"))
      await self._send_command(bytes.fromhex("aa01192842"))  # it starts moving again

    await _reset_to_zero()

    # 7 - wait for home position to change
    # go_to_bucket{1,2} does not work until the home position changes
    start = await self.get_home_position()
    num_tries = 0
    while await self.get_home_position() == start:
      await asyncio.sleep(0.1)
      num_tries += 1
      if num_tries % 25 == 0:
        await _reset_to_zero()
      if num_tries > 100:
        raise RuntimeError("Home position did not change after spin.")


# Deprecated alias with warning # TODO: remove mid May 2025 (giving people 1 month to update)
# https://github.com/PyLabRobot/pylabrobot/issues/466


class VSpin:
  def __init__(self, *args, **kwargs):
    raise RuntimeError("`VSpin` is deprecated. Please use `VSpinBackend` instead. ")
